"""DataLoaders for Anomaly Tasks."""
# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Optional

import pandas as pd
from bson import ObjectId

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset


class BaseAnomalyDataset(DatasetEntity, ABC):
    """Base Dataloader for Anomaly Tasks."""

    def __init__(
        self,
        train_subset: Optional[Dict[str, str]] = None,
        val_subset: Optional[Dict[str, str]] = None,
        test_subset: Optional[Dict[str, str]] = None,
    ):
        """Base Anomaly Dataset.

        Args:
            train_subset (Optional[Dict[str, str]], optional): Path to annotation
                and dataset used for training. Defaults to None.
            val_subset (Optional[Dict[str, str]], optional): Path to annotation
                and dataset used for validation. Defaults to None.
            test_subset (Optional[Dict[str, str]], optional): Path to annotation
                and dataset used for testing. Defaults to None.
        """
        items: List[DatasetItemEntity] = []
        self.normal_label = LabelEntity(id=ID(0), name="Normal", domain=Domain.ANOMALY_CLASSIFICATION)
        self.abnormal_label = LabelEntity(
            id=ID(1),
            name="Anomalous",
            domain=Domain.ANOMALY_CLASSIFICATION,
            is_anomalous=True,
        )

        if train_subset is not None:
            train_ann_file = Path(train_subset["ann_file"])
            train_data_root = Path(train_subset["data_root"])
            items.extend(
                self.get_dataset_items(
                    ann_file_path=train_ann_file,
                    data_root_dir=train_data_root,
                    subset=Subset.TRAINING,
                )
            )

        if val_subset is not None:
            val_ann_file = Path(val_subset["ann_file"])
            val_data_root = Path(val_subset["data_root"])
            items.extend(
                self.get_dataset_items(
                    ann_file_path=val_ann_file,
                    data_root_dir=val_data_root,
                    subset=Subset.VALIDATION,
                )
            )

        if test_subset is not None:
            test_ann_file = Path(test_subset["ann_file"])
            test_data_root = Path(test_subset["data_root"])
            items.extend(
                self.get_dataset_items(
                    ann_file_path=test_ann_file,
                    data_root_dir=test_data_root,
                    subset=Subset.TESTING,
                )
            )

        super().__init__(items=items)

    @abstractmethod
    def get_dataset_items(self, ann_file_path: Path, data_root_dir: Path, subset: Subset) -> List[DatasetItemEntity]:
        """To be implemented ib subclasses."""
        raise NotImplementedError


class AnomalyClassificationDataset(BaseAnomalyDataset):
    """Dataloader for Anomaly Classification Task.

    Example:
    >>> train_subset = {
            "ann_file": "tests/assets/anomaly/classification/train.json",
            "data_root": "tests/assets/anomaly/shapes",
        }
    >>> val_subset = {
            "ann_file": "tests/assets/anomaly/classification/val.json",
            "data_root": "tests/assets/anomaly/shapes"
        }
    >>> training_dataset = AnomalyClassificationDataset(
            train_subset=train_subset, val_subset=val_subset
        )
    >>> test_subset = {
            "ann_file": "tests/assets/anomaly/classification/test.json",
            "data_root": "tests/assets/anomaly/shapes"
        }
    >>> testing_dataset = AnomalyClassificationDataset(test_subset=test_subset)
    """

    def get_dataset_items(self, ann_file_path: Path, data_root_dir: Path, subset: Subset) -> List[DatasetItemEntity]:
        """Loads dataset based on the image path in annotation file.

        Args:
            ann_file_path (Path): Path to json containing the annotations.
                For example of annotation look at `tests/assets/anomaly/[train, test,val].json.
            data_root_dir (Path): Path to folder containing images.
            subset (Subset): Subset of the dataset.

        Returns:
            List[DatasetItemEntity]: List containing subset dataset.
        """
        # read annotation file
        samples = pd.read_json(ann_file_path)

        dataset_items = []
        for _, sample in samples.iterrows():
            # Create image
            # convert path to str as PosixPath is not supported by Image
            image = Image(file_path=str(data_root_dir / sample.image_path))
            # Create annotation
            shape = Rectangle.generate_full_box()
            label: LabelEntity = self.normal_label if sample.label == "good" else self.abnormal_label
            labels = [ScoredLabel(label, probability=1.0)]
            annotations = [Annotation(shape=shape, labels=labels)]
            annotation_scene = AnnotationSceneEntity(annotations=annotations, kind=AnnotationSceneKind.ANNOTATION)

            # Create dataset item
            dataset_item = DatasetItemEntity(media=image, annotation_scene=annotation_scene, subset=subset)
            # Add to dataset items
            dataset_items.append(dataset_item)

        return dataset_items


class AnomalySegmentationDataset(BaseAnomalyDataset):
    """Dataloader for Anomaly Segmentation Task.

    Example:
        >>> train_subset = {
                "ann_file": "tests/assets/anomaly/segmentation/train.json",
                "data_root": "tests/assets/anomaly/shapes",
            }
        >>> val_subset = {
                "ann_file": "tests/assets/anomaly/segmentation/val.json",
                "data_root": "tests/assets/anomaly/shapes"
            }
        >>> training_dataset = AnomalySegmentationDataset(
                train_subset=train_subset, val_subset=val_subset
            )
        >>> test_subset = {
                "ann_file": "tests/assets/anomaly/segmentation/test.json",
                "data_root": "tests/assets/anomaly/shapes"
            }
        >>> testing_dataset = AnomalySegmentationDataset(test_subset=test_subset)

    """

    def get_dataset_items(self, ann_file_path: Path, data_root_dir: Path, subset: Subset) -> List[DatasetItemEntity]:
        """Loads dataset based on the image path in annotation file.

        Args:
            ann_file_path (Path): Path to json containing the annotations.
                For example of annotation look at `tests/assets/anomaly/[train, test,val].json.
            data_root_dir (Path): Path to folder containing images.
            subset (Subset): Subset of the dataset.

        Returns:
            List[DatasetItemEntity]: List containing subset dataset.
        """
        # read annotation file
        samples = pd.read_json(ann_file_path)

        dataset_items = []
        for _, sample in samples.iterrows():
            # Create image
            # convert path to str as PosixPath is not supported by Image
            image = Image(file_path=str(data_root_dir / sample.image_path))
            # Create annotation
            label: LabelEntity = self.normal_label if sample.label == "good" else self.abnormal_label
            annotations = [
                Annotation(
                    Rectangle.generate_full_box(),
                    labels=[ScoredLabel(label=label, probability=1.0)],
                )
            ]
            if isinstance(sample.masks, list) and len(sample.masks) > 0:
                for contour in sample.masks:
                    points = [Point(x, y) for x, y in contour]
                    polygon = Polygon(points=points)
                    if polygon.get_area() > 0:
                        # Contour is a closed polygon with area > 0
                        annotations.append(
                            Annotation(
                                shape=polygon,
                                labels=[ScoredLabel(label, 1.0)],
                                id=ID(ObjectId()),
                            )
                        )
                    else:
                        # Contour is a closed polygon with area == 0
                        warnings.warn(
                            "The geometry of the segmentation map you are converting "
                            "is not fully supported. Polygons with a area of zero "
                            "will be removed.",
                            UserWarning,
                        )
            annotation_scene = AnnotationSceneEntity(annotations=annotations, kind=AnnotationSceneKind.ANNOTATION)

            # Add to dataset items
            dataset_items.append(DatasetItemEntity(media=image, annotation_scene=annotation_scene, subset=subset))

        return dataset_items


class AnomalyDetectionDataset(BaseAnomalyDataset):
    """Dataloader for Anomaly Segmentation Task.

    Example:
        >>> train_subset = {
                "ann_file": "tests/assets/anomaly/detection/train.json",
                "data_root": "tests/assets/anomaly/shapes",
            }
        >>> val_subset = {
                "ann_file": "tests/assets/anomaly/detection/val.json",
                "data_root": "tests/assets/anomaly/shapes"
            }
        >>> training_dataset = AnomalyDetectionDataset(
                train_subset=train_subset, val_subset=val_subset
            )
        >>> test_subset = {
                "ann_file": "tests/assets/anomaly/detection/test.json",
                "data_root": "tests/assets/anomaly/shapes"
            }
        >>> testing_dataset = AnomalyDetectionDataset(test_subset=test_subset)

    """

    def get_dataset_items(self, ann_file_path: Path, data_root_dir: Path, subset: Subset) -> List[DatasetItemEntity]:
        """Loads dataset based on the image path in annotation file.

        Args:
            ann_file_path (Path): Path to json containing the annotations.
                For example of annotation look at `tests/assets/anomaly/[train, test,val].json.
            data_root_dir (Path): Path to folder containing images.
            subset (Subset): Subset of the dataset.

        Returns:
            List[DatasetItemEntity]: List containing subset dataset.
        """
        # read annotation file
        samples = pd.read_json(ann_file_path)

        dataset_items = []
        for _, sample in samples.iterrows():
            # Create image
            # convert path to str as PosixPath is not supported by Image
            image = Image(file_path=str(data_root_dir / sample.image_path))
            # Create annotation
            label: LabelEntity = self.normal_label if sample.label == "good" else self.abnormal_label
            annotations = [
                Annotation(
                    Rectangle.generate_full_box(),
                    labels=[ScoredLabel(label=label, probability=1.0)],
                )
            ]
            if isinstance(sample.bboxes, list) and len(sample.bboxes) > 0:
                for bbox in sample.bboxes:
                    box = Rectangle(bbox[0], bbox[1], bbox[2], bbox[3])
                    if box.get_area() > 0:
                        # Contour is a closed polygon with area > 0
                        annotations.append(
                            Annotation(
                                shape=box,
                                labels=[ScoredLabel(label, 1.0)],
                                id=ID(ObjectId()),
                            )
                        )
                    else:
                        # Contour is a closed polygon with area == 0
                        warnings.warn(
                            "The geometry of the segmentation map you are converting "
                            "is not fully supported. Polygons with a area of zero "
                            "will be removed.",
                            UserWarning,
                        )
            annotation_scene = AnnotationSceneEntity(annotations=annotations, kind=AnnotationSceneKind.ANNOTATION)

            # Add to dataset items
            dataset_items.append(DatasetItemEntity(media=image, annotation_scene=annotation_scene, subset=subset))

        return dataset_items



"""Sample Code of otx training for action detection."""

# Copyright (C) 2021-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import argparse
import sys

from mmcv.utils import get_logger

from otx.algorithms.common.utils import get_task_class
from otx.api.configuration.helper import create
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.core.data.adapter import get_dataset_adapter

logger = get_logger(name="sample")


def parse_args():
    """Parse function for getting model template & check export."""
    parser = argparse.ArgumentParser(description="Sample showcasing the new API")
    parser.add_argument("template_file_path", help="path to template file")
    parser.add_argument("--export", action="store_true")
    return parser.parse_args()


TRAIN_DATA_ROOTS = "tests/assets/cvat_dataset/action_detection/train"
VAL_DATA_ROOTS = "tests/assets/cvat_dataset/action_detection/train"


def load_test_dataset(model_template):
    """Load Sample dataset for detection."""
    dataset_adapter = get_dataset_adapter(
        model_template.task_type,
        model_template.train_type,
        train_data_roots=TRAIN_DATA_ROOTS,
        val_data_roots=VAL_DATA_ROOTS,
    )
    dataset = dataset_adapter.get_otx_dataset()
    label_schema = dataset_adapter.get_label_schema()
    return dataset, label_schema


# pylint: disable=too-many-locals, too-many-statements
def main(args):
    """Main function of Detection Sample."""
    logger.info("Fine tuning sample dataset")
    logger.info("Sample dataset can be found at tests/assets/cvat_dataset/action_classification/train")

    logger.info("Load model template")
    model_template = parse_model_template(args.template_file_path)

    logger.info("Get dataset")
    dataset, labels_schema = load_test_dataset(model_template)

    logger.info("Set hyperparameters")
    params = create(model_template.hyper_parameters.data)
    params.learning_parameters.num_iters = 5

    logger.info("Setup environment")
    environment = TaskEnvironment(
        model=None,
        hyper_parameters=params,
        label_schema=labels_schema,
        model_template=model_template,
    )

    logger.info("Create base Task")
    task_impl_path = model_template.entrypoints.base
    task_cls = get_task_class(task_impl_path)
    task = task_cls(task_environment=environment)

    logger.info("Train model")
    output_model = ModelEntity(
        dataset,
        environment.get_model_configuration(),
    )
    task.train(dataset, output_model)

    logger.info("Get predictions on the validation set")
    validation_dataset = dataset.get_subset(Subset.VALIDATION)
    predicted_validation_dataset = task.infer(
        validation_dataset.with_empty_annotations(),
        InferenceParameters(is_evaluation=False),
    )
    resultset = ResultSetEntity(
        model=output_model,
        ground_truth_dataset=validation_dataset,
        prediction_dataset=predicted_validation_dataset,
    )
    logger.info("Estimate quality on validation set")
    task.evaluate(resultset)
    logger.info(str(resultset.performance))

    if args.export:
        raise Exception("CVS-102941 ONNX export of action detection model keeps failed")


if __name__ == "__main__":
    sys.exit(main(parse_args()) or 0)



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv import ConfigDict

from otx.mpa.cls.stage import ClsStage, Stage
from otx.mpa.utils.config_utils import update_or_add_custom_hook
from otx.mpa.utils.logger import get_logger

logger = get_logger()

CLASS_INC_DATASET = [
    "OTXClsDataset",
    "OTXMultilabelClsDataset",
    "MPAHierarchicalClsDataset",
    "ClsTVDataset",
]
PSEUDO_LABEL_ENABLE_DATASET = ["ClassIncDataset", "ClsTVDataset"]
WEIGHT_MIX_CLASSIFIER = ["SAMImageClassifier"]


class IncrClsStage(ClsStage):
    """Patch config to support incremental learning for object Cls"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def configure_task(self, cfg, training, **kwargs):
        """Patch config to support incremental learning"""
        super().configure_task(cfg, training, **kwargs)
        if "task_adapt" in cfg:
            self.configure_task_adapt(cfg, training, **kwargs)

    # noqa: C901
    def configure_task_adapt(self, cfg, training, **kwargs):
        """Configure for Task Adaptation Task"""

        self.adapt_type = cfg["task_adapt"].get("op", "REPLACE")
        train_data_cfg = Stage.get_data_cfg(cfg, "train")
        if training:
            if train_data_cfg.type not in CLASS_INC_DATASET:
                logger.warning(f"Class Incremental Learning for {train_data_cfg.type} is not yet supported!")
            if "new_classes" not in train_data_cfg:
                logger.warning('"new_classes" should be defined for incremental learning w/ current model.')

            if cfg.model.type in WEIGHT_MIX_CLASSIFIER:
                cfg.model.task_adapt = ConfigDict(
                    src_classes=self.model_classes,
                    dst_classes=self.data_classes,
                )
            else:
                logger.warning(f"Weight mixing for {cfg.model.type} is not yet supported!")

            # refine self.dst_class following adapt_type (REPLACE, MERGE)
            self.refine_classes(train_data_cfg)
            cfg.model.head.num_classes = len(self.dst_classes)

            # configure loss, sampler, task_adapt_hook
            self.configure_task_modules(cfg)

        else:  # if eval phase (eval)
            if train_data_cfg.get("new_classes"):
                self.refine_classes(train_data_cfg)
                cfg.model.head.num_classes = len(self.dst_classes)

    def configure_task_modules(self, cfg):
        if not cfg.model.get("multilabel", False) and not cfg.model.get("hierarchical", False):
            efficient_mode = cfg["task_adapt"].get("efficient_mode", True)
            sampler_type = "balanced"
            self.configure_loss(cfg)
        else:
            efficient_mode = cfg["task_adapt"].get("efficient_mode", False)
            sampler_type = "cls_incr"

        if len(set(self.model_classes) & set(self.dst_classes)) == 0 or set(self.model_classes) == set(
            self.dst_classes
        ):
            sampler_flag = False
        else:
            sampler_flag = True

        # Update Task Adapt Hook
        task_adapt_hook = ConfigDict(
            type="TaskAdaptHook",
            src_classes=self.old_classes,
            dst_classes=self.dst_classes,
            model_type=cfg.model.type,
            sampler_flag=sampler_flag,
            sampler_type=sampler_type,
            efficient_mode=efficient_mode,
        )
        update_or_add_custom_hook(cfg, task_adapt_hook)

    def configure_loss(self, cfg):
        if len(set(self.model_classes) & set(self.dst_classes)) == 0 or set(self.model_classes) == set(
            self.dst_classes
        ):
            cfg.model.head.loss = dict(type="CrossEntropyLoss", loss_weight=1.0)
        else:
            cfg.model.head.loss = ConfigDict(
                type="IBLoss",
                num_classes=cfg.model.head.num_classes,
            )
            ib_loss_hook = ConfigDict(
                type="IBLossHook",
                dst_classes=self.dst_classes,
            )
            update_or_add_custom_hook(cfg, ib_loss_hook)

    def refine_classes(self, train_cfg):
        # Get 'new_classes' in data.train_cfg & get 'old_classes' pretreained model meta data CLASSES
        new_classes = train_cfg["new_classes"]
        self.old_classes = self.model_meta["CLASSES"]
        if self.adapt_type == "REPLACE":
            # if 'REPLACE' operation, then self.dst_classes -> data_classes
            self.dst_classes = self.data_classes.copy()
        elif self.adapt_type == "MERGE":
            # if 'MERGE' operation, then self.dst_classes -> old_classes + new_classes (merge)
            self.dst_classes = self.old_classes + [cls for cls in new_classes if cls not in self.old_classes]
        else:
            raise KeyError(f"{self.adapt_type} is not supported for task_adapt options!")
        train_cfg.classes = self.dst_classes



"""This module contains the types used for specifying UI interaction with the configuration."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from enum import Enum, auto


class Operator(Enum):
    """This Enum represents the allowed operators for use in constructing UI rules for configuration parameters.

    These operators can be used to disable a configuration parameter, conditional on the value of another parameter.
    """

    NOT = auto()
    EQUAL_TO = auto()
    LESS_THAN = auto()
    GREATER_THAN = auto()
    AND = auto()
    OR = auto()
    XOR = auto()

    def __str__(self):
        """Retrieves the string representation of an instance of the Enum."""
        return self.name


class Action(Enum):
    """This Enum represents the allowed actions that UI rules can dictate on configuration parameters."""

    HIDE = auto()
    SHOW = auto()
    ENABLE_EDITING = auto()
    DISABLE_EDITING = auto()

    def __str__(self):
        """Retrieves the string representation of an instance of the Enum."""
        return self.name



img_size = (992, 736)
img_norm_cfg = dict(mean=[0, 0, 0], std=[255, 255, 255], to_rgb=True)

train_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="LoadAnnotations", with_bbox=True),
    dict(type="MinIoURandomCrop", min_ious=(0.1, 0.3, 0.5, 0.7, 0.9), min_crop_size=0.3),
    dict(
        type="Resize",
        img_scale=[(992, 736), (896, 736), (1088, 736), (992, 672), (992, 800)],
        multiscale_mode="value",
        keep_ratio=False,
    ),
    dict(type="RandomFlip", flip_ratio=0.5),
    dict(type="Normalize", **img_norm_cfg),
    dict(type="DefaultFormatBundle"),
    dict(type="Collect", keys=["img", "gt_bboxes", "gt_labels"]),
]
test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(
        type="MultiScaleFlipAug",
        img_scale=img_size,
        flip=False,
        transforms=[
            dict(type="Resize", keep_ratio=False),
            dict(type="RandomFlip"),
            dict(type="Normalize", **img_norm_cfg),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ],
    ),
]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Dict, List, Optional

from otx.mpa.utils.logger import get_logger

from ..builder import PARSERS
from ..parser import parameter_parser

logger = get_logger()


NECK_INPUT_TYPES = ["ReduceMean", "MaxPool", "AvgPool"]
NECK_TYPES = [
    "Reshape",
    "Squeeze",
    "Unsqueeze",
    "Concat",
    "Convert",
    "ShapeOf",
    "StridedSlice",
    "Transpose",
]


@PARSERS.register()
def cls_base_parser(graph, component: str = "backbone") -> Optional[Dict[str, List[str]]]:
    assert component in ["backbone", "neck", "head"]

    result_nodes = graph.get_nodes_by_types(["Result"])
    if len(result_nodes) != 1:
        logger.debug("More than one reulst nodes are found.")
        return None
    result_node = result_nodes[0]

    neck_input = None
    for _, node_to in graph.bfs(result_node, True, 20):
        if node_to.type in NECK_INPUT_TYPES:
            logger.debug(f"Found neck_input: {node_to.name}")
            neck_input = node_to
            break

    if neck_input is None:
        logger.debug("Can not determine the output of backbone.")
        return None

    neck_output = neck_input
    for node_from, node_to in graph.bfs(neck_input, False, 10):
        done = False
        for node_to_ in node_to:
            if node_to_.type not in NECK_TYPES:
                done = True
                break
        neck_output = node_from
        if done:
            break

    if component == "backbone":
        outputs = [node.name for node in graph.predecessors(neck_input) if node.type != "Constant"]
        if len(outputs) != 1:
            logger.debug(f"neck_input {neck_input.name} has more than one predecessors.")
            return None

        inputs = parameter_parser(graph)
        if len(inputs) != 1:
            logger.debug("More than on parameter nodes are found.")
            return None

        return dict(
            inputs=inputs,
            outputs=outputs,
        )

    elif component == "neck":
        return dict(
            inputs=[neck_input.name],
            outputs=[neck_output.name],
        )

    elif component == "head":
        inputs = list(graph.successors(neck_output))
        #  if len(inputs) != 1:
        #      logger.debug(f"neck_output {neck_output.name} has more than one successors.")
        #      return None

        outputs = graph.get_nodes_by_types(["Result"])
        if len(outputs) != 1:
            logger.debug("more than one network output are found.")
            return None
        for node_from, node_to in graph.bfs(outputs[0], True, 5):
            if node_to.type == "Softmax":
                outputs = [node_from]
                break

        if not graph.has_path(inputs[0], outputs[0]):
            logger.debug(f"input({inputs[0].name}) and output({outputs[0].name}) are reversed")
            return None

        return dict(
            inputs=[input.name for input in inputs],
            outputs=[output.name for output in outputs],
        )



"""OpenVINO Training Extensions."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

__version__ = "1.0.0"
# NOTE: Sync w/ otx/api/usecases/exportable_code/demo/requirements.txt



"""Sample Code of otx training for detection."""

# Copyright (C) 2021-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import argparse
import sys

import numpy as np
from mmcv.utils import get_logger

from otx.algorithms.common.utils import get_task_class
from otx.api.configuration.helper import create
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType

logger = get_logger(name="mmdet")


def parse_args():
    """Parse function for getting model template & check export."""
    parser = argparse.ArgumentParser(description="Sample showcasing the new API")
    parser.add_argument("template_file_path", help="path to template file")
    parser.add_argument("--export", action="store_true")
    return parser.parse_args()


colors = dict(red=(255, 0, 0), green=(0, 255, 0))


def load_test_dataset(data_type):
    """Load Sample dataset for detection."""

    def gen_image(resolution, x1, y1, x2, y2, color):
        width, height = resolution
        image = np.full([height, width, 3], fill_value=255, dtype=np.uint8)
        image[int(y1 * height) : int(y2 * height), int(x1 * width) : int(x2 * width), :] = np.full(
            [int(height * (y2 - y1)), int(width * (x2 - x1)), 3],
            fill_value=colors[color],
            dtype=np.uint8,
        )
        return (image, Rectangle(x1=x1, y1=y1, x2=x2, y2=y2))

    images = [
        (0.0, 0.0, 0.5, 0.5),
        (0.5, 0.0, 1.0, 0.5),
        (0.0, 0.5, 0.5, 1.0),
        (0.5, 0.5, 1.0, 1.0),
    ]
    labels = [
        LabelEntity(name="red", domain=Domain.DETECTION, id=0),  # OLD class
        LabelEntity(name="green", domain=Domain.DETECTION, id=1),
    ]

    def get_image(i, subset, label_id):
        image, bbox = gen_image((640, 480), *images[i], labels[label_id].name)
        return DatasetItemEntity(
            media=Image(data=image),
            annotation_scene=AnnotationSceneEntity(
                annotations=[Annotation(bbox, labels=[ScoredLabel(label=labels[label_id])])],
                kind=AnnotationSceneKind.ANNOTATION,
            ),
            subset=subset,
        )

    old_train = [
        get_image(0, Subset.TRAINING, 0),
        get_image(1, Subset.TRAINING, 0),
        get_image(2, Subset.TRAINING, 0),
        get_image(3, Subset.TRAINING, 0),
        get_image(0, Subset.TRAINING, 0),
        get_image(1, Subset.TRAINING, 0),
        get_image(2, Subset.TRAINING, 0),
        get_image(3, Subset.TRAINING, 0),
        get_image(0, Subset.TRAINING, 0),
        get_image(1, Subset.TRAINING, 0),
    ]
    old_val = [
        get_image(0, Subset.VALIDATION, 0),
        get_image(1, Subset.VALIDATION, 0),
        get_image(2, Subset.VALIDATION, 0),
        get_image(3, Subset.VALIDATION, 0),
        get_image(0, Subset.TESTING, 0),
        get_image(1, Subset.TESTING, 0),
        get_image(2, Subset.TESTING, 0),
        get_image(3, Subset.TESTING, 0),
    ]
    new_train = [
        get_image(0, Subset.TRAINING, 0),
        get_image(1, Subset.TRAINING, 0),
        get_image(2, Subset.TRAINING, 0),
        get_image(3, Subset.TRAINING, 0),
        get_image(0, Subset.TRAINING, 1),
        get_image(1, Subset.TRAINING, 1),
        get_image(2, Subset.TRAINING, 1),
        get_image(3, Subset.TRAINING, 1),
        get_image(0, Subset.TRAINING, 1),
        get_image(1, Subset.TRAINING, 1),
    ]
    new_val = [
        get_image(0, Subset.VALIDATION, 0),
        get_image(1, Subset.VALIDATION, 0),
        get_image(2, Subset.VALIDATION, 0),
        get_image(3, Subset.VALIDATION, 0),
        get_image(1, Subset.VALIDATION, 1),
        get_image(2, Subset.VALIDATION, 1),
        get_image(3, Subset.VALIDATION, 1),
        get_image(0, Subset.TESTING, 0),
        get_image(1, Subset.TESTING, 0),
        get_image(2, Subset.TESTING, 0),
        get_image(3, Subset.TESTING, 0),
        get_image(1, Subset.TESTING, 1),
        get_image(2, Subset.TESTING, 1),
        get_image(3, Subset.TESTING, 1),
    ]
    old = old_train + old_val
    new = new_train + new_val
    if data_type == "old":
        return DatasetEntity(old), [labels[0]]
    return DatasetEntity(old + new), labels


# pylint: disable=too-many-locals, too-many-statements
def main(args):
    """Main function of Detection Sample."""
    logger.info("[SL] Train initial model with OLD dataset")
    dataset, labels_list = load_test_dataset("old")
    labels_schema = LabelSchemaEntity.from_labels(labels_list)

    logger.info(f"Train dataset: {len(dataset.get_subset(Subset.TRAINING))} items")
    logger.info(f"Validation dataset: {len(dataset.get_subset(Subset.VALIDATION))} items")

    logger.info("Load model template")
    model_template = parse_model_template(args.template_file_path)

    logger.info("Set hyperparameters")
    params = create(model_template.hyper_parameters.data)
    params.learning_parameters.num_iters = 5
    params.learning_parameters.learning_rate = 0.01
    params.learning_parameters.learning_rate_warmup_iters = 1
    params.learning_parameters.batch_size = 4

    logger.info("Setup environment")
    environment = TaskEnvironment(
        model=None,
        hyper_parameters=params,
        label_schema=labels_schema,
        model_template=model_template,
    )

    logger.info("Create base Task")
    task_impl_path = model_template.entrypoints.base
    task_cls = get_task_class(task_impl_path)
    task = task_cls(task_environment=environment)

    logger.info("Train model")
    initial_model = ModelEntity(
        dataset,
        environment.get_model_configuration(),
    )
    task.train(dataset, initial_model)

    logger.info("Class-incremental learning with OLD + NEW dataset")
    dataset, labels_list = load_test_dataset("new")
    labels_schema = LabelSchemaEntity.from_labels(labels_list)

    logger.info(f"Train dataset: {len(dataset.get_subset(Subset.TRAINING))} items")
    logger.info(f"Validation dataset: {len(dataset.get_subset(Subset.VALIDATION))} items")

    logger.info("Load model template")
    model_template = parse_model_template(args.template_file_path)

    logger.info("Set hyperparameters")
    params = create(model_template.hyper_parameters.data)
    params.learning_parameters.num_iters = 5
    params.learning_parameters.learning_rate = 0.01
    params.learning_parameters.learning_rate_warmup_iters = 1
    params.learning_parameters.batch_size = 4

    logger.info("Setup environment")
    environment = TaskEnvironment(
        model=initial_model,
        hyper_parameters=params,
        label_schema=labels_schema,
        model_template=model_template,
    )

    logger.info("Create base Task")
    task_impl_path = model_template.entrypoints.base
    task_cls = get_task_class(task_impl_path)
    task = task_cls(task_environment=environment)

    logger.info("Train model")
    output_model = ModelEntity(
        dataset,
        environment.get_model_configuration(),
    )
    task.train(dataset, output_model)

    logger.info("Get predictions on the validation set")
    validation_dataset = dataset.get_subset(Subset.VALIDATION)
    predicted_validation_dataset = task.infer(
        validation_dataset.with_empty_annotations(),
        InferenceParameters(is_evaluation=True),
    )
    resultset = ResultSetEntity(
        model=output_model,
        ground_truth_dataset=validation_dataset,
        prediction_dataset=predicted_validation_dataset,
    )
    logger.info("Estimate quality on validation set")
    task.evaluate(resultset)
    logger.info(str(resultset.performance))

    if args.export:
        logger.info("Export model")
        exported_model = ModelEntity(
            dataset,
            environment.get_model_configuration(),
        )
        task.export(ExportType.OPENVINO, exported_model)

        logger.info("Create OpenVINO Task")
        environment.model = exported_model
        openvino_task_impl_path = model_template.entrypoints.openvino
        openvino_task_cls = get_task_class(openvino_task_impl_path)
        openvino_task = openvino_task_cls(environment)

        logger.info("Get predictions on the validation set")
        predicted_validation_dataset = openvino_task.infer(
            validation_dataset.with_empty_annotations(),
            InferenceParameters(is_evaluation=True),
        )

        resultset = ResultSetEntity(
            model=output_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )
        logger.info("Estimate quality on validation set")
        openvino_task.evaluate(resultset)
        logger.info(str(resultset.performance))

        logger.info("Run POT optimization")
        optimized_model = ModelEntity(
            dataset,
            environment.get_model_configuration(),
        )
        openvino_task.optimize(
            OptimizationType.POT,
            dataset,
            optimized_model,
            OptimizationParameters(),
        )

        logger.info("Get predictions on the validation set")
        predicted_validation_dataset = openvino_task.infer(
            validation_dataset.with_empty_annotations(),
            InferenceParameters(is_evaluation=True),
        )
        resultset = ResultSetEntity(
            model=optimized_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )
        logger.info("Performance of optimized model:")
        openvino_task.evaluate(resultset)
        logger.info(str(resultset.performance))

        logger.info("Running the NNCF optimization")
        environment.model = output_model
        nncf_task_impl_path = model_template.entrypoints.nncf
        nncf_task_cls = get_task_class(nncf_task_impl_path)
        nncf_task = nncf_task_cls(environment)

        optimized_model = ModelEntity(
            dataset,
            configuration=environment.get_model_configuration(),
        )
        nncf_task.optimize(OptimizationType.NNCF, dataset, optimized_model)

        logger.info("Inferring the optimised model on the validation set.")
        predicted_validation_dataset = nncf_task.infer(
            validation_dataset.with_empty_annotations(),
            InferenceParameters(is_evaluation=True),
        )
        resultset = ResultSetEntity(
            model=optimized_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )

        logger.info("Evaluating the optimized model on the validation set.")
        nncf_task.evaluate(resultset)
        logger.info(str(resultset.performance))

        logger.info("Exporting the model.")
        exported_model = ModelEntity(
            train_dataset=dataset,
            configuration=environment.get_model_configuration(),
        )
        nncf_task.export(ExportType.OPENVINO, exported_model)
        environment.model = exported_model

        logger.info("Creating the OpenVINO Task.")
        environment.model = exported_model
        openvino_task_impl_path = model_template.entrypoints.openvino
        nncf_openvino_task_cls = get_task_class(openvino_task_impl_path)
        nncf_openvino_task = nncf_openvino_task_cls(environment)

        logger.info("Inferring the exported model on the validation set.")
        predicted_validation_dataset = nncf_openvino_task.infer(
            validation_dataset.with_empty_annotations(),
            InferenceParameters(is_evaluation=True),
        )

        logger.info("Evaluating the exported model on the validation set.")
        resultset = ResultSetEntity(
            model=exported_model,
            ground_truth_dataset=validation_dataset,
            prediction_dataset=predicted_validation_dataset,
        )
        nncf_openvino_task.evaluate(resultset)
        logger.info(str(resultset.performance))


if __name__ == "__main__":
    sys.exit(main(parse_args()) or 0)



__img_scale = (992, 736)
__img_norm_cfg = dict(mean=[0, 0, 0], std=[255, 255, 255], to_rgb=True)

common_pipeline = [
    dict(
        type="Resize",
        img_scale=[
            (992, 736),
            (896, 736),
            (1088, 736),
            (992, 672),
            (992, 800),
        ],
        multiscale_mode="value",
        keep_ratio=False,
    ),
    dict(type="RandomFlip", flip_ratio=0.5),
    dict(type="BranchImage", key_map=dict(img="img0")),
    dict(type="NDArrayToPILImage", keys=["img"]),
    dict(
        type="RandomApply",
        transform_cfgs=[
            dict(
                type="ColorJitter",
                brightness=0.4,
                contrast=0.4,
                saturation=0.4,
                hue=0.1,
            )
        ],
        p=0.8,
    ),
    dict(type="RandomGrayscale", p=0.2),
    dict(
        type="RandomApply",
        transform_cfgs=[
            dict(
                type="RandomGaussianBlur",
                sigma_min=0.1,
                sigma_max=2.0,
            )
        ],
        p=0.5,
    ),
    dict(type="PILImageToNDArray", keys=["img"]),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="Pad", size_divisor=32),
    dict(type="NDArrayToTensor", keys=["img", "img0"]),
    dict(
        type="RandomErasing",
        p=0.7,
        scale=[0.05, 0.2],
        ratio=[0.3, 3.3],
        value="random",
    ),
    dict(
        type="RandomErasing",
        p=0.5,
        scale=[0.02, 0.2],
        ratio=[0.10, 6.0],
        value="random",
    ),
    dict(
        type="RandomErasing",
        p=0.3,
        scale=[0.02, 0.2],
        ratio=[0.05, 8.0],
        value="random",
    ),
]

train_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="LoadAnnotations", with_bbox=True),
    dict(type="MinIoURandomCrop", min_ious=(0.1, 0.3, 0.5, 0.7, 0.9), min_crop_size=0.3),
    *common_pipeline,
    dict(type="ToTensor", keys=["gt_bboxes", "gt_labels"]),
    dict(
        type="ToDataContainer",
        fields=[
            dict(key="img", stack=True),
            dict(key="img0", stack=True),
            dict(key="gt_bboxes"),
            dict(key="gt_labels"),
        ],
    ),
    dict(
        type="Collect",
        keys=["img", "img0", "gt_bboxes", "gt_labels"],
    ),
]

unlabeled_pipeline = [
    dict(type="LoadImageFromFile"),
    *common_pipeline,
    dict(
        type="ToDataContainer",
        fields=[
            dict(key="img", stack=True),
            dict(key="img0", stack=True),
        ],
    ),
    dict(
        type="Collect",
        keys=[
            "img",
            "img0",
        ],
    ),
]

test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(
        type="MultiScaleFlipAug",
        img_scale=__img_scale,
        flip=False,
        transforms=[
            dict(type="Resize", keep_ratio=False),
            dict(type="Normalize", **__img_norm_cfg),
            dict(type="Pad", size_divisor=32),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ],
    ),
]



"""Collection of utils for data in Detection Task."""

# Copyright (C) 2021-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
import math
import os.path as osp
from typing import Any, Dict, List, Optional, Sequence

import numpy as np
from mmdet.datasets.api_wrappers.coco_api import COCO

from otx.algorithms.detection.adapters.mmdet.data.dataset import (
    get_annotation_mmdet_format,
)
from otx.algorithms.detection.configs.base.configuration import DetectionConfig
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    DirectoryPathCheck,
    JsonFilePathCheck,
    OptionalDirectoryPathCheck,
    check_input_parameters_type,
)
from otx.api.utils.shape_factory import ShapeFactory


# pylint: disable=too-many-instance-attributes, too-many-arguments
@check_input_parameters_type({"path": JsonFilePathCheck})
def get_classes_from_annotation(path):
    """Return classes from annotation."""
    with open(path, encoding="UTF-8") as read_file:
        content = json.load(read_file)
        categories = [v["name"] for v in sorted(content["categories"], key=lambda x: x["id"])]
    return categories


class LoadAnnotations:
    """Load Annotations class."""

    @check_input_parameters_type()
    def __init__(self, with_bbox: bool = True, with_label: bool = True, with_mask: bool = False):
        self.with_bbox = with_bbox
        self.with_label = with_label
        self.with_mask = with_mask

    @staticmethod
    def _load_bboxes(results):
        ann_info = results["ann_info"]
        results["gt_bboxes"] = ann_info["bboxes"].copy()

        gt_bboxes_ignore = ann_info.get("bboxes_ignore", None)
        if gt_bboxes_ignore is not None:
            results["gt_bboxes_ignore"] = gt_bboxes_ignore.copy()
            results["bbox_fields"].append("gt_bboxes_ignore")
        results["bbox_fields"].append("gt_bboxes")
        return results

    @staticmethod
    def _load_labels(results):
        results["gt_labels"] = results["ann_info"]["labels"].copy()
        return results

    @staticmethod
    def _load_masks(results):
        gt_masks = results["ann_info"]["masks"]
        results["gt_masks"] = gt_masks
        results["mask_fields"].append("gt_masks")
        return results

    @check_input_parameters_type()
    def __call__(self, results: Dict[str, Any]):
        """Callback function of LoadAnnotations."""
        if self.with_bbox:
            results = LoadAnnotations._load_bboxes(results)
            if results is None:
                return None
        if self.with_label:
            results = LoadAnnotations._load_labels(results)
        if self.with_mask:
            results = LoadAnnotations._load_masks(results)

        return results

    def __repr__(self):
        """String function of LoadAnnotations."""
        repr_str = self.__class__.__name__
        repr_str += f"(with_bbox={self.with_bbox}, "
        repr_str += f"with_label={self.with_label})"
        return repr_str


class CocoDataset:
    """CocoDataset."""

    @check_input_parameters_type({"ann_file": JsonFilePathCheck, "data_root": OptionalDirectoryPathCheck})
    def __init__(
        self,
        ann_file: str,
        classes: Optional[Sequence[str]] = None,
        data_root: Optional[str] = None,
        img_prefix: str = "",
        test_mode: bool = False,
        filter_empty_gt: bool = True,
        min_size: Optional[int] = None,
        with_mask: bool = False,
    ):
        self.ann_file = ann_file
        self.data_root = data_root
        self.img_prefix = img_prefix
        self.test_mode = test_mode
        self.filter_empty_gt = filter_empty_gt
        self.classes = self.get_classes(classes)
        self.min_size = min_size
        self.with_mask = with_mask

        if self.data_root is not None:
            # if not osp.isabs(self.ann_file):
            #     self.ann_file = osp.join(self.data_root, self.ann_file)
            if not (self.img_prefix is None or osp.isabs(self.img_prefix)):
                self.img_prefix = osp.join(self.data_root, self.img_prefix)

        self.data_infos = self.load_annotations(self.ann_file)

        if not test_mode:
            valid_inds = self._filter_imgs()
            self.data_infos = [self.data_infos[i] for i in valid_inds]

    def __len__(self):
        """Length of CocoDataset."""
        return len(self.data_infos)

    @check_input_parameters_type()
    def pre_pipeline(self, results: Dict[str, Any]):
        """Initialize pipeline."""
        results["img_prefix"] = self.img_prefix
        results["bbox_fields"] = []
        results["mask_fields"] = []
        results["seg_fields"] = []

    def _rand_another(self, idx):
        """Get Random indices."""
        pool = np.where(self.flag == self.flag[idx])[0]
        return np.random.choice(pool)

    @check_input_parameters_type()
    def __getitem__(self, idx: int):
        """Return dataset item from index."""
        return self.prepare_img(idx)

    def __iter__(self):
        """Iterator of CocoDataset."""
        for i in range(len(self)):
            yield self[i]

    @check_input_parameters_type()
    def prepare_img(self, idx: int):
        """Load Annotations function with images."""
        img_info = self.data_infos[idx]
        ann_info = self.get_ann_info(idx)
        results = dict(img_info=img_info, ann_info=ann_info)
        self.pre_pipeline(results)
        return LoadAnnotations(with_mask=self.with_mask)(results)

    @check_input_parameters_type()
    def get_classes(self, classes: Optional[Sequence[str]] = None):
        """Return classes function."""
        if classes is None:
            return get_classes_from_annotation(self.ann_file)

        if isinstance(classes, (tuple, list)):
            return classes

        raise ValueError(f"Unsupported type {type(classes)} of classes.")

    @check_input_parameters_type({"ann_file": JsonFilePathCheck})
    def load_annotations(self, ann_file):
        """Load annotations function from coco."""
        self.coco = COCO(ann_file)
        self.cat_ids = self.coco.get_cat_ids(cat_names=self.classes)
        self.cat2label = {cat_id: i for i, cat_id in enumerate(self.cat_ids)}
        self.img_ids = self.coco.get_img_ids()
        data_infos = []
        for i in self.img_ids:
            info = self.coco.load_imgs([i])[0]
            info["filename"] = info["file_name"]
            data_infos.append(info)
        return data_infos

    @check_input_parameters_type()
    def get_ann_info(self, idx: int):
        """Getting Annotation info."""
        img_id = self.data_infos[idx]["id"]
        ann_ids = self.coco.get_ann_ids(img_ids=[img_id])
        ann_info = self.coco.load_anns(ann_ids)
        return self._parse_ann_info(self.data_infos[idx], ann_info)

    @check_input_parameters_type()
    def get_cat_ids(self, idx: int):
        """Getting cat_ids."""
        img_id = self.data_infos[idx]["id"]
        ann_ids = self.coco.get_ann_ids(img_ids=[img_id])
        ann_info = self.coco.load_anns(ann_ids)
        return [ann["category_id"] for ann in ann_info]

    def _filter_imgs(self, min_size=32):
        """Filter images too small or without ground truths."""
        valid_inds = []
        # obtain images that contain annotation
        ids_with_ann = set(_["image_id"] for _ in self.coco.anns.values())
        # obtain images that contain annotations of the required categories
        ids_in_cat = set()
        for i, class_id in enumerate(self.cat_ids):
            ids_in_cat |= set(self.coco.cat_img_map[class_id])
        # merge the image id sets of the two conditions and use the merged set
        # to filter out images if self.filter_empty_gt=True
        ids_in_cat &= ids_with_ann

        valid_img_ids = []
        for i, img_info in enumerate(self.data_infos):
            img_id = self.img_ids[i]
            if self.filter_empty_gt and img_id not in ids_in_cat:
                continue
            if min(img_info["width"], img_info["height"]) >= min_size:
                valid_inds.append(i)
                valid_img_ids.append(img_id)
        self.img_ids = valid_img_ids
        return valid_inds

    def _parse_ann_info(self, img_info, ann_info):  # pylint: disable=too-many-locals, too-many-branches
        """Parse annotation info."""
        gt_bboxes = []
        gt_labels = []
        gt_bboxes_ignore = []
        gt_masks_ann = []
        for ann in ann_info:
            if ann.get("ignore", False):
                continue
            x1, y1, width, height = ann["bbox"]
            inter_w = max(0, min(x1 + width, img_info["width"]) - max(x1, 0))
            inter_h = max(0, min(y1 + height, img_info["height"]) - max(y1, 0))
            if inter_w * inter_h == 0:
                continue
            if ann["area"] <= 0 or width < 1 or height < 1:
                continue
            if self.min_size is not None:
                if width < self.min_size or height < self.min_size:
                    continue
            if ann["category_id"] not in self.cat_ids:
                continue
            bbox = [x1, y1, x1 + width, y1 + height]
            if ann.get("iscrowd", False):
                gt_bboxes_ignore.append(bbox)
            else:
                gt_bboxes.append(bbox)
                gt_labels.append(self.cat2label[ann["category_id"]])
                gt_masks_ann.append(ann.get("segmentation", None))

        if gt_bboxes:
            gt_bboxes = np.array(gt_bboxes, dtype=np.float32)
            gt_labels = np.array(gt_labels, dtype=np.int64)
        else:
            gt_bboxes = np.zeros((0, 4), dtype=np.float32)
            gt_labels = np.array([], dtype=np.int64)

        if gt_bboxes_ignore:
            gt_bboxes_ignore = np.array(gt_bboxes_ignore, dtype=np.float32)
        else:
            gt_bboxes_ignore = np.zeros((0, 4), dtype=np.float32)

        seg_map = img_info["filename"].replace("jpg", "png")

        ann = dict(
            bboxes=gt_bboxes,
            labels=gt_labels,
            bboxes_ignore=gt_bboxes_ignore,
            masks=gt_masks_ann,
            seg_map=seg_map,
        )

        return ann


@check_input_parameters_type()
def find_label_by_name(labels: List[LabelEntity], name: str, domain: Domain):
    """Return label from name."""
    matching_labels = [label for label in labels if label.name == name]
    if len(matching_labels) == 1:
        return matching_labels[0]
    if len(matching_labels) == 0:
        label = LabelEntity(name=name, domain=domain, id=ID(len(labels)))
        labels.append(label)
        return label
    raise ValueError("Found multiple matching labels")


@check_input_parameters_type({"ann_file_path": JsonFilePathCheck, "data_root_dir": DirectoryPathCheck})
def load_dataset_items_coco_format(
    ann_file_path: str,
    data_root_dir: str,
    domain: Domain,
    subset: Subset = Subset.NONE,
    labels_list: Optional[List[LabelEntity]] = None,
    with_mask: bool = False,
):  # pylint: disable=too-many-locals
    """Load dataset from CocoDataset."""
    test_mode = subset in {Subset.VALIDATION, Subset.TESTING}

    coco_dataset = CocoDataset(
        ann_file=ann_file_path,
        data_root=data_root_dir,
        classes=None,
        test_mode=test_mode,
        with_mask=with_mask,
    )
    coco_dataset.test_mode = False
    for label_name in coco_dataset.classes:
        find_label_by_name(labels_list, label_name, domain)

    dataset_items = []
    for item in coco_dataset:

        def create_gt_box(x1, y1, x2, y2, label_name):
            return Annotation(
                Rectangle(x1=x1, y1=y1, x2=x2, y2=y2),
                labels=[ScoredLabel(label=find_label_by_name(labels_list, label_name, domain))],
            )

        def create_gt_polygon(polygon_group, label_name):
            if len(polygon_group) != 1:
                raise RuntimeError(
                    "Complex instance segmentation masks consisting of several polygons are not supported."
                )

            return Annotation(
                Polygon(points=polygon_group[0]),
                labels=[ScoredLabel(label=find_label_by_name(labels_list, label_name, domain))],
            )

        img_height = item["img_info"].get("height")
        img_width = item["img_info"].get("width")
        divisor = np.array(
            [img_width, img_height, img_width, img_height],
            dtype=item["gt_bboxes"].dtype,
        )
        bboxes = item["gt_bboxes"] / divisor
        labels = item["gt_labels"]

        assert len(bboxes) == len(labels)
        if with_mask:
            polygons = item["gt_masks"]
            assert len(bboxes) == len(polygons)
            normalized_polygons = []  # type: List
            for polygon_group in polygons:
                normalized_polygons.append([])
                for polygon in polygon_group:
                    normalized_polygon = [p / divisor[i % 2] for i, p in enumerate(polygon)]
                    points = [
                        Point(normalized_polygon[i], normalized_polygon[i + 1]) for i in range(0, len(polygon), 2)
                    ]
                    normalized_polygons[-1].append(points)

        if item["img_prefix"] is not None:
            filename = osp.join(item["img_prefix"], item["img_info"]["filename"])
        else:
            filename = item["img_info"]["filename"]

        if with_mask:
            shapes = [
                create_gt_polygon(polygon_group, coco_dataset.classes[label_id])
                for polygon_group, label_id in zip(normalized_polygons, labels)
            ]
        else:
            shapes = [
                create_gt_box(x1, y1, x2, y2, coco_dataset.classes[label_id])
                for (x1, y1, x2, y2), label_id in zip(bboxes, labels)
            ]

        dataset_item = DatasetItemEntity(
            media=Image(file_path=filename),
            annotation_scene=AnnotationSceneEntity(annotations=shapes, kind=AnnotationSceneKind.ANNOTATION),
            subset=subset,
        )
        dataset_items.append(dataset_item)

    return dataset_items


@check_input_parameters_type({"dataset": DatasetParamTypeCheck})
def get_sizes_from_dataset_entity(dataset: DatasetEntity, target_wh: List[int]):
    """Function to get sizes of instances in DatasetEntity and to resize it to the target size.

    :param dataset: DatasetEntity in which to get statistics
    :param target_wh: target width and height of the dataset
    :return list: tuples with width and height of each instance
    """
    wh_stats = []
    for item in dataset:
        for ann in item.get_annotations(include_empty=False):
            has_detection_labels = any(
                label.domain == Domain.DETECTION for label in ann.get_labels(include_empty=False)
            )
            if has_detection_labels:
                box = ShapeFactory.shape_as_rectangle(ann.shape)
                width = box.width * target_wh[0]
                height = box.height * target_wh[1]
                wh_stats.append((width, height))
    return wh_stats


@check_input_parameters_type()
def get_anchor_boxes(wh_stats: List[tuple], group_as: List[int]):
    """Get anchor box widths & heights."""
    from sklearn.cluster import KMeans

    kmeans = KMeans(init="k-means++", n_clusters=sum(group_as), random_state=0).fit(wh_stats)
    centers = kmeans.cluster_centers_

    areas = np.sqrt(np.prod(centers, axis=1))
    idx = np.argsort(areas)

    widths = centers[idx, 0]
    heights = centers[idx, 1]

    group_as = np.cumsum(group_as[:-1])
    widths, heights = np.split(widths, group_as), np.split(heights, group_as)
    widths = [width.tolist() for width in widths]
    heights = [height.tolist() for height in heights]
    return widths, heights


@check_input_parameters_type()
def format_list_to_str(value_lists: list):
    """Decrease floating point digits in logs."""
    str_value = ""
    for value_list in value_lists:
        str_value += "[" + ", ".join(f"{value:.2f}" for value in value_list) + "], "
    return f"[{str_value[:-2]}]"


# TODO [Eugene] please add unit test for this function
def adaptive_tile_params(
    tiling_parameters: DetectionConfig.BaseTilingParameters, dataset: DatasetEntity, object_tile_ratio=0.01, rule="avg"
):
    """Config tile parameters.

    Adapt based on annotation statistics.
    i.e. tile size, tile overlap, ratio and max objects per sample

    Args:
        tiling_parameters (BaseTilingParameters): tiling parameters of the model
        dataset (DatasetEntity): training dataset
        object_tile_ratio (float, optional): The desired ratio of object area and tile area. Defaults to 0.01.
        rule (str, optional): min or avg.  In `min` mode, tile size is computed based on the smallest object, and in
                              `avg` mode tile size is computed by averaging all the object areas. Defaults to "avg".

    """
    assert rule in ["min", "avg"], f"Unknown rule: {rule}"

    bboxes = np.zeros((0, 4), dtype=np.float32)
    labels = dataset.get_labels(include_empty=False)
    domain = labels[0].domain
    max_object = 0
    for dataset_item in dataset:
        result = get_annotation_mmdet_format(dataset_item, labels, domain)
        if len(result["bboxes"]):
            bboxes = np.concatenate((bboxes, result["bboxes"]), 0)
            if len(result["bboxes"]) > max_object:
                max_object = len(result["bboxes"])

    areas = (bboxes[:, 2] - bboxes[:, 0]) * (bboxes[:, 3] - bboxes[:, 1])

    if rule == "min":
        object_area = np.min(areas)
    elif rule == "avg":
        object_area = np.mean(areas)
    max_area = np.max(areas)

    tile_size = int(math.sqrt(object_area / object_tile_ratio))
    tile_overlap = max_area / (tile_size**2)

    # validate parameters are in range
    tile_size = max(
        tiling_parameters.get_metadata("tile_size")["min_value"],
        min(tiling_parameters.get_metadata("tile_size")["max_value"], tile_size),
    )
    tile_overlap = max(
        tiling_parameters.get_metadata("tile_overlap")["min_value"],
        min(tiling_parameters.get_metadata("tile_overlap")["max_value"], tile_overlap),
    )
    max_object = max(
        tiling_parameters.get_metadata("tile_max_number")["min_value"],
        min(tiling_parameters.get_metadata("tile_max_number")["max_value"], max_object),
    )

    tiling_parameters.tile_size = tile_size
    tiling_parameters.tile_max_number = max_object
    tiling_parameters.tile_overlap = tile_overlap



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from copy import deepcopy

import numpy as np
from mmcls.datasets.builder import PIPELINES
from mmcls.datasets.pipelines import Compose, to_tensor
from mmcv.utils import build_from_cfg


@PIPELINES.register_module()
class TwoCropTransform:
    """Generate two different cropped views of an image"""

    def __init__(self, pipeline):
        self.pipeline1 = Compose([build_from_cfg(p, PIPELINES) for p in pipeline])
        self.pipeline2 = Compose([build_from_cfg(p, PIPELINES) for p in pipeline])

    def __call__(self, data):
        data1 = self.pipeline1(deepcopy(data))
        data2 = self.pipeline2(deepcopy(data))

        data = deepcopy(data1)
        data["img"] = to_tensor(np.ascontiguousarray(np.stack((data1["img"], data2["img"]), axis=0)))
        return data



"""NNCF utils."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from mmcv import Config
from mmcv.parallel import DataContainer, collate, scatter
from torch import nn
from torch.utils.data import DataLoader

from otx.algorithms.common.adapters.mmcv.nncf.runners import NNCF_META_KEY
from otx.algorithms.common.adapters.mmcv.utils.builder import build_data_parallel
from otx.algorithms.common.adapters.nncf.compression import (
    is_checkpoint_nncf,
    is_state_nncf,
)
from otx.algorithms.common.adapters.nncf.utils import (
    check_nncf_is_enabled,
    load_checkpoint,
    no_nncf_trace,
)
from otx.algorithms.common.utils import get_arg_spec
from otx.mpa.utils.logger import get_logger

logger = get_logger()


def get_fake_input(
    preprocessor: Callable[..., Dict[str, Any]],
    data: Optional[np.ndarray] = None,
    shape: Tuple[int, ...] = (128, 128, 3),
    device: Union[str, torch.device] = "cpu",
):
    """A function to generate fake data."""

    if isinstance(device, str):
        device = torch.device(device)

    if data is None:
        data = dict(img=np.zeros(shape, dtype=np.uint8))
    else:
        data = dict(img=data)
    data = preprocessor(data)

    for key, value in data.items():
        if not isinstance(value, list):
            data[key] = [value]

    if device.type == "cpu":
        data = scatter(collate([data], samples_per_gpu=1), [-1])[0]
    elif device.type == "cuda":
        data = scatter(collate([data], samples_per_gpu=1), [device.index])[0]
    else:
        raise NotImplementedError()
    return data


def model_eval(
    model: nn.Module,
    *,
    config: Config,
    val_dataloader: DataLoader,
    evaluate_fn: Callable,
    distributed: bool,
):
    """A model evaluation function for NNCF.

    Runs evaluation of the model on the validation set and
    returns the target metric value.
    Used to evaluate the original model before compression
    if NNCF-based accuracy-aware training is used.
    """
    if val_dataloader is None:
        raise RuntimeError(
            "Cannot perform model evaluation on the validation "
            "dataset since the validation data loader was not passed "
            "to wrap_nncf_model"
        )

    nncf_config = config.get("nncf_config")
    metric_name = nncf_config.get("target_metric_name")
    prepared_model = build_data_parallel(model, config, distributed=distributed)

    logger.info("Calculating an original model accuracy")

    evaluation_cfg = deepcopy(config.evaluation)
    spec = get_arg_spec(val_dataloader.dataset.evaluate)
    for key in list(evaluation_cfg.keys()):
        if key not in spec:
            evaluation_cfg.pop(key)
    evaluation_cfg["metric"] = metric_name

    if distributed:  # pylint: disable=no-else-return
        dist_eval_res: List[Dict[str, Any]] = [{}]
        results = evaluate_fn(prepared_model, val_dataloader, gpu_collect=True)
        if torch.distributed.get_rank() == 0:
            eval_res = val_dataloader.dataset.evaluate(results, **evaluation_cfg)
            if metric_name not in eval_res:
                raise RuntimeError(f"Cannot find {metric_name} metric in the evaluation result dict")
            dist_eval_res[0] = eval_res

        torch.distributed.broadcast_object_list(dist_eval_res, src=0)
        return dist_eval_res[0][metric_name]
    else:
        results = evaluate_fn(prepared_model, val_dataloader, show=False)
        eval_res = val_dataloader.dataset.evaluate(results, **evaluation_cfg)

        if metric_name not in eval_res:
            raise RuntimeError(f"Cannot find {metric_name} metric in the evaluation result dict {eval_res.keys()}")

        return eval_res[metric_name]


# pylint: disable-next=too-many-branches,too-many-statements,too-many-locals
def wrap_nncf_model(  # noqa: C901
    config: Config,
    model: nn.Module,
    *,
    model_eval_fn: Optional[Callable] = None,
    dummy_forward_fn: Optional[Callable] = None,
    get_fake_input_fn: Optional[Callable] = None,
    wrap_inputs_fn: Optional[Callable] = None,
    dataloader_for_init: Optional[DataLoader] = None,
    init_state_dict: Optional[Dict[Any, Any]] = None,
    is_accuracy_aware: bool = False,
):
    """The function wraps mmcv model by NNCF."""

    check_nncf_is_enabled()

    from nncf import NNCFConfig
    from nncf.torch import (
        create_compressed_model,
        load_state,
        register_default_init_args,
    )
    from nncf.torch.dynamic_graph.io_handling import nncf_model_input
    from nncf.torch.initialization import PTInitializingDataLoader

    class _MMInitializeDataLoader(PTInitializingDataLoader):
        def get_inputs(self, dataloader_output):
            # redefined PTInitializingDataLoader because
            # of DataContainer format in mmdet
            kwargs = {k: v.data[0] if isinstance(v, DataContainer) else v for k, v in dataloader_output.items()}
            return (), kwargs

    nncf_config = NNCFConfig(config.nncf_config)
    resuming_state_dict = None

    if dataloader_for_init:
        wrapped_loader = _MMInitializeDataLoader(dataloader_for_init)
        eval_fn = model_eval_fn if is_accuracy_aware else None
        nncf_config = register_default_init_args(
            nncf_config,
            wrapped_loader,
            model_eval_fn=eval_fn,
            device=next(model.parameters()).device,
        )

    if config.get("resume_from"):
        checkpoint_path = config.get("resume_from")
        assert is_checkpoint_nncf(checkpoint_path), (
            "It is possible to resume training with NNCF compression from NNCF checkpoints only. "
            'Use "load_from" with non-compressed model for further compression by NNCF.'
        )
    elif config.get("load_from"):
        checkpoint_path = config.get("load_from")
        if not is_checkpoint_nncf(checkpoint_path):
            logger.info("Received non-NNCF checkpoint to start training -- initialization of NNCF fields will be done")
    else:
        checkpoint_path = None

    if not dataloader_for_init and not checkpoint_path and not init_state_dict:
        logger.warning(
            "Either dataloader_for_init or NNCF pre-trained "
            "model checkpoint should be set. Without this, "
            "quantizers will not be initialized"
        )

    if init_state_dict:
        assert is_state_nncf(init_state_dict)
        meta_state = init_state_dict["meta"][NNCF_META_KEY]
        resuming_state_dict = init_state_dict["state_dict"]
        compression_state = meta_state.compression_ctrl
    elif checkpoint_path:
        logger.info(f"Loading NNCF checkpoint from {checkpoint_path}")
        logger.info(
            "Please, note that this first loading is made before addition of "
            "NNCF FakeQuantize nodes to the model, so there may be some "
            "warnings on unexpected keys"
        )
        compression_state, resuming_state_dict = load_checkpoint(model, checkpoint_path)
        logger.info(f"Loaded NNCF checkpoint from {checkpoint_path}")
    else:
        compression_state = None

    if dummy_forward_fn is None:
        assert get_fake_input_fn is not None

        def _get_fake_data_for_forward(nncf_config):
            device = next(model.parameters()).device

            if nncf_config.get("input_info", None) and nncf_config.get("input_info").get("sample_size", None):
                input_size = nncf_config.get("input_info").get("sample_size")
                assert len(input_size) == 4 and input_size[0] == 1
                H, W, C = input_size[2], input_size[3], input_size[1]  # pylint: disable=invalid-name
                shape = tuple([H, W, C])
            else:
                shape = (128, 128, 3)

            with no_nncf_trace():
                return get_fake_input_fn(shape=shape, device=device)

        def dummy_forward_fn(model):
            fake_data = _get_fake_data_for_forward(nncf_config)
            img, img_metas = fake_data["img"], fake_data["img_metas"]

            ctx = model.nncf_trace_context(img_metas)
            with ctx:
                # The device where model is could be changed under this context
                img = [i.to(next(model.parameters()).device) for i in img]
                # Marking data as NNCF network input must be after device movement
                img = [nncf_model_input(i) for i in img]
                model(img)

    if wrap_inputs_fn is None:

        def wrap_inputs_fn(args, kwargs):
            img = kwargs.get("img") if "img" in kwargs else args[0]
            if isinstance(img, list):
                assert len(img) == 1, "Input list must have a length 1"
                assert torch.is_tensor(img[0]), "Input for a model must be a tensor"
                img[0] = nncf_model_input(img[0])
            else:
                assert torch.is_tensor(img), "Input for a model must be a tensor"
                img = nncf_model_input(img)
            if "img" in kwargs:
                kwargs["img"] = img
            else:
                args = (img, *args[1:])
            return args, kwargs

    if "log_dir" in nncf_config:
        os.makedirs(nncf_config["log_dir"], exist_ok=True)

    compression_ctrl, model = create_compressed_model(
        model,
        nncf_config,
        dummy_forward_fn=dummy_forward_fn,
        wrap_inputs_fn=wrap_inputs_fn,
        compression_state=compression_state,
    )

    if resuming_state_dict:
        load_state(model, resuming_state_dict, is_resume=True)

    return compression_ctrl, model



from typing import Any, Dict

import numpy as np
import pytest
from PIL import Image

from otx.algorithms.segmentation.adapters.mmseg.data.pipelines import (
    NDArrayToPILImage,
    PILImageToNDArray,
    RandomColorJitter,
    RandomGaussianBlur,
    RandomGrayscale,
    RandomResizedCrop,
    RandomSolarization,
    TwoCropTransform,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@pytest.fixture(scope="module")
def inputs_np():
    return {
        "img": np.random.randint(0, 10, (16, 16, 3), dtype=np.uint8),
        "gt_semantic_seg": np.random.rand(16, 16),
        "flip": True,
    }


@pytest.fixture(scope="module")
def inputs_PIL():
    return {
        "img": Image.fromarray(np.random.randint(0, 10, (16, 16, 3), dtype=np.uint8)),
        "gt_semantic_seg": np.random.randint(0, 5, (16, 16), dtype=np.uint8),
        "seg_fields": ["gt_semantic_seg"],
        "ori_shape": (16, 16, 3),
    }


class TestTwoCropTransform:
    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        mocker.patch(
            "otx.algorithms.segmentation.adapters.mmseg.data.pipelines.build_from_cfg", return_value=lambda x: x
        )
        self.two_crop_transform = TwoCropTransform(view0=[], view1=[])

    @e2e_pytest_unit
    def test_call(self, mocker, inputs_np: Dict[str, Any]) -> None:
        """Test __call__."""
        results = self.two_crop_transform(inputs_np)

        assert isinstance(results, dict)
        assert "img" in results and results["img"].ndim == 4
        assert "gt_semantic_seg" in results and results["gt_semantic_seg"].ndim == 3
        assert "flip" in results and isinstance(results["flip"], list)

    @e2e_pytest_unit
    def test_call_with_single_pipeline(self, mocker, inputs_np: Dict[str, Any]) -> None:
        """Test __call__ with single pipeline."""
        self.two_crop_transform.is_both = False

        results = self.two_crop_transform(inputs_np)

        assert isinstance(results, dict)
        assert "img" in results and results["img"].ndim == 3
        assert "gt_semantic_seg" in results and results["gt_semantic_seg"].ndim == 2
        assert "flip" in results and isinstance(results["flip"], bool)


@e2e_pytest_unit
def test_random_resized_crop(inputs_PIL: Dict[str, Any]) -> None:
    """Test RandomResizedCrop."""
    random_resized_crop = RandomResizedCrop(size=(8, 8))

    results = random_resized_crop(inputs_PIL)

    assert isinstance(results, dict)
    assert "img" in results and results["img"].size == (8, 8)
    assert "gt_semantic_seg" in results and results["gt_semantic_seg"].shape == (8, 8)
    assert "img_shape" in results
    assert "ori_shape" in results
    assert "scale_factor" in results


@e2e_pytest_unit
def test_random_color_jitter(inputs_PIL: Dict[str, Any]) -> None:
    """Test RandomColorJitter."""
    random_color_jitter = RandomColorJitter(p=1.0)

    results = random_color_jitter(inputs_PIL)

    assert isinstance(results, dict)
    assert "img" in results


@e2e_pytest_unit
def test_random_grayscale(inputs_PIL: Dict[str, Any]) -> None:
    """Test RandomGrayscale."""
    random_grayscale = RandomGrayscale()

    results = random_grayscale(inputs_PIL)

    assert isinstance(results, dict)
    assert "img" in results


@e2e_pytest_unit
def test_random_gaussian_blur(inputs_PIL: Dict[str, Any]) -> None:
    """Test RandomGaussianBlur."""
    random_gaussian_blur = RandomGaussianBlur(p=1.0, kernel_size=3)

    results = random_gaussian_blur(inputs_PIL)

    assert isinstance(results, dict)
    assert "img" in results


@e2e_pytest_unit
def test_random_solarization(inputs_np: Dict[str, Any]) -> None:
    """Test RandomSolarization."""
    random_solarization = RandomSolarization(p=1.0)

    results = random_solarization(inputs_np)

    assert isinstance(results, dict)
    assert "img" in results
    assert repr(random_solarization) == "RandomSolarization"


@e2e_pytest_unit
def test_nd_array_to_pil_image(inputs_np: Dict[str, Any]) -> None:
    """Test NDArrayToPILImage."""
    nd_array_to_pil_image = NDArrayToPILImage(keys=["img"])

    results = nd_array_to_pil_image(inputs_np)

    assert "img" in results
    assert isinstance(results["img"], Image.Image)
    assert repr(nd_array_to_pil_image) == "NDArrayToPILImage"


@e2e_pytest_unit
def test_pil_image_to_nd_array(inputs_PIL: Dict[str, Any]) -> None:
    """Test PILImageToNDArray."""
    pil_image_to_nd_array = PILImageToNDArray(keys=["img"])

    results = pil_image_to_nd_array(inputs_PIL)

    assert "img" in results
    assert isinstance(results["img"], np.ndarray)
    assert repr(pil_image_to_nd_array) == "PILImageToNDArray"



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from collections import OrderedDict
from functools import partial

from mmcls.models.builder import CLASSIFIERS
from mmcls.models.classifiers.base import BaseClassifier
from mmcls.models.classifiers.image import ImageClassifier

from otx.mpa.deploy.utils import is_mmdeploy_enabled
from otx.mpa.modules.utils.task_adapt import map_class_names
from otx.mpa.utils.logger import get_logger

from .sam_classifier_mixin import SAMClassifierMixin

logger = get_logger()


@CLASSIFIERS.register_module()
class SAMImageClassifier(SAMClassifierMixin, ImageClassifier):
    """SAM-enabled ImageClassifier"""

    def __init__(self, task_adapt=None, **kwargs):
        if "multilabel" in kwargs:
            self.multilabel = kwargs.pop("multilabel")
        if "hierarchical" in kwargs:
            self.hierarchical = kwargs.pop("hierarchical")
        super().__init__(**kwargs)
        self.is_export = False
        # Hooks for redirect state_dict load/save
        self._register_state_dict_hook(self.state_dict_hook)
        self._register_load_state_dict_pre_hook(partial(self.load_state_dict_pre_hook, self))
        if task_adapt:
            self._register_load_state_dict_pre_hook(
                partial(
                    self.load_state_dict_mixing_hook,
                    self,  # model
                    task_adapt["dst_classes"],  # model_classes
                    task_adapt["src_classes"],  # chkpt_classes
                )
            )

    def forward_train(self, img, gt_label, **kwargs):
        """Forward computation during training.

        Args:
            img (Tensor): of shape (N, C, H, W) encoding input images.
                Typically these should be mean centered and std scaled.

            gt_label (Tensor): It should be of shape (N, 1) encoding the
                ground-truth label of input images for single label task. It
                shoulf be of shape (N, C) encoding the ground-truth label
                of input images for multi-labels task.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """
        if self.augments is not None:
            img, gt_label = self.augments(img, gt_label)

        x = self.extract_feat(img)

        losses = dict()

        if self.multilabel or self.hierarchical:
            loss = self.head.forward_train(x, gt_label, **kwargs)
        else:
            gt_label = gt_label.squeeze(dim=1)
            loss = self.head.forward_train(x, gt_label)

        losses.update(loss)

        return losses

    @staticmethod
    def state_dict_hook(module, state_dict, prefix, *args, **kwargs):
        """Redirect model as output state_dict for OTX model compatibility"""
        backbone_type = type(module.backbone).__name__
        if backbone_type not in ["OTXMobileNetV3", "OTXEfficientNet", "OTXEfficientNetV2"]:
            return

        if backbone_type == "OTXMobileNetV3":
            for k in list(state_dict.keys()):
                v = state_dict.pop(k)
                if not prefix or k.startswith(prefix):
                    k = k.replace(prefix, "", 1)
                    if k.startswith("backbone"):
                        k = k.replace("backbone.", "", 1)
                    elif k.startswith("head"):
                        k = k.replace("head.", "", 1)
                        if "3" in k:  # MPA uses "classifier.3", OTX uses "classifier.4". Convert for OTX compatibility.
                            k = k.replace("3", "4")
                            if module.multilabel and not module.is_export:
                                v = v.t()
                    k = prefix + k
                state_dict[k] = v

        elif backbone_type == "OTXEfficientNet":
            for k in list(state_dict.keys()):
                v = state_dict.pop(k)
                if not prefix or k.startswith(prefix):
                    k = k.replace(prefix, "", 1)
                    if k.startswith("backbone"):
                        k = k.replace("backbone.", "", 1)
                    elif k.startswith("head"):
                        k = k.replace("head", "output", 1)
                        if not module.hierarchical and not module.is_export:
                            k = k.replace("fc", "asl")
                            v = v.t()
                    k = prefix + k
                state_dict[k] = v

        elif backbone_type == "OTXEfficientNetV2":
            for k in list(state_dict.keys()):
                v = state_dict.pop(k)
                if not prefix or k.startswith(prefix):
                    k = k.replace(prefix, "", 1)
                    if k.startswith("backbone"):
                        k = k.replace("backbone.", "", 1)
                    elif k == "head.fc.weight":
                        k = k.replace("head.fc", "model.classifier")
                        if not module.hierarchical and not module.is_export:
                            v = v.t()
                    k = prefix + k
                state_dict[k] = v

        return state_dict

    @staticmethod
    def load_state_dict_pre_hook(module, state_dict, prefix, *args, **kwargs):
        """Redirect input state_dict to model for OTX model compatibility"""
        backbone_type = type(module.backbone).__name__
        if backbone_type not in ["OTXMobileNetV3", "OTXEfficientNet", "OTXEfficientNetV2"]:
            return

        if backbone_type == "OTXMobileNetV3":
            for k in list(state_dict.keys()):
                v = state_dict.pop(k)
                if not prefix or k.startswith(prefix):
                    k = k.replace(prefix, "", 1)
                    if k.startswith("classifier."):
                        if "4" in k:
                            k = "head." + k.replace("4", "3")
                            if module.multilabel:
                                v = v.t()
                        else:
                            k = "head." + k
                    elif k.startswith("act"):
                        k = "head." + k
                    elif not k.startswith("backbone."):
                        k = "backbone." + k
                    k = prefix + k
                state_dict[k] = v

        elif backbone_type == "OTXEfficientNet":
            for k in list(state_dict.keys()):
                v = state_dict.pop(k)
                if not prefix or k.startswith(prefix):
                    k = k.replace(prefix, "", 1)
                    if k.startswith("features.") and "activ" not in k:
                        k = "backbone." + k
                    elif k.startswith("output."):
                        k = k.replace("output", "head")
                        if not module.hierarchical:
                            k = k.replace("asl", "fc")
                            v = v.t()
                    k = prefix + k
                state_dict[k] = v

        elif backbone_type == "OTXEfficientNetV2":
            for k in list(state_dict.keys()):
                v = state_dict.pop(k)
                if not prefix or k.startswith(prefix):
                    k = k.replace(prefix, "", 1)
                    if k.startswith("model.classifier"):
                        k = k.replace("model.classifier", "head.fc")
                        if not module.hierarchical:
                            v = v.t()
                    elif k.startswith("model"):
                        k = "backbone." + k
                    k = prefix + k
                state_dict[k] = v
        else:
            logger.info("conversion is not required.")

    @staticmethod
    def load_state_dict_mixing_hook(model, model_classes, chkpt_classes, chkpt_dict, prefix, *args, **kwargs):
        """Modify input state_dict according to class name matching before weight loading"""
        backbone_type = type(model.backbone).__name__
        if backbone_type not in ["OTXMobileNetV3", "OTXEfficientNet", "OTXEfficientNetV2"]:
            return

        # Dst to src mapping index
        model_classes = list(model_classes)
        chkpt_classes = list(chkpt_classes)
        model2chkpt = map_class_names(model_classes, chkpt_classes)
        logger.info(f"{chkpt_classes} -> {model_classes} ({model2chkpt})")
        model_dict = model.state_dict()

        if backbone_type == "OTXMobileNetV3":
            if model.multilabel:
                param_names = ["classifier.4.weight"]
            else:
                param_names = ["classifier.4.weight", "classifier.4.bias"]

        elif backbone_type == "OTXEfficientNet":
            if not model.hierarchical:
                param_names = ["output.asl.weight"]
            else:
                param_names = ["output.fc.weight"]

        elif backbone_type == "OTXEfficientNetV2":
            param_names = [
                "model.classifier.weight",
            ]
            if "head.fc.bias" in chkpt_dict.keys():
                param_names.append("head.fc.bias")

        for model_name in param_names:
            model_param = model_dict[model_name].clone()
            if backbone_type == "OTXMobileNetV3":
                chkpt_name = "head." + model_name.replace("4", "3")
                if model.multilabel:
                    model_param = model_param.t()
            elif backbone_type in "OTXEfficientNet":
                chkpt_name = model_name.replace("output", "head")
                if not model.hierarchical:
                    chkpt_name = chkpt_name.replace("asl", "fc")
                    model_param = model_param.t()

            elif backbone_type in "OTXEfficientNetV2":
                if model_name.endswith("bias"):
                    chkpt_name = model_name
                else:
                    chkpt_name = model_name.replace("model.classifier", "head.fc")
                    if not model.hierarchical:
                        model_param = model_param.t()

            if model_name not in model_dict or chkpt_name not in chkpt_dict:
                logger.info(f"Skipping weight copy: {chkpt_name}")
                continue

            # Mix weights
            chkpt_param = chkpt_dict[chkpt_name]
            for m, c in enumerate(model2chkpt):
                if c >= 0:
                    model_param[m].copy_(chkpt_param[c])

            # Replace checkpoint weight by mixed weights
            chkpt_dict[chkpt_name] = model_param

    def extract_feat(self, img):
        """Directly extract features from the backbone + neck
        Overriding for OpenVINO export with features
        """
        x = self.backbone(img)
        # For Global Backbones (det/seg/etc..),
        # In case of tuple or list, only the feat of the last layer is used.
        if isinstance(x, (tuple, list)):
            x = x[-1]

        if self.with_neck:
            x = self.neck(x)

        return x


if is_mmdeploy_enabled():
    from mmdeploy.core import FUNCTION_REWRITER

    from otx.mpa.modules.hooks.recording_forward_hooks import (
        FeatureVectorHook,
        ReciproCAMHook,
    )

    @FUNCTION_REWRITER.register_rewriter(
        "otx.mpa.modules.models.classifiers.sam_classifier.SAMImageClassifier.extract_feat"
    )
    def sam_image_classifier__extract_feat(ctx, self, img):
        feat = self.backbone(img)
        # For Global Backbones (det/seg/etc..),
        # In case of tuple or list, only the feat of the last layer is used.
        if isinstance(feat, (tuple, list)):
            feat = feat[-1]
        backbone_feat = feat
        if self.with_neck:
            feat = self.neck(feat)
        return feat, backbone_feat

    @FUNCTION_REWRITER.register_rewriter(
        "otx.mpa.modules.models.classifiers.sam_classifier.SAMImageClassifier.simple_test"
    )
    def sam_image_classifier__simple_test(ctx, self, img, img_metas):
        feat, backbone_feat = self.extract_feat(img)
        logit = self.head.simple_test(feat)

        if ctx.cfg["dump_features"]:
            saliency_map = ReciproCAMHook(self).func(backbone_feat)
            feature_vector = FeatureVectorHook.func(backbone_feat)
            return logit, feature_vector, saliency_map

        return logit



"""Model wrapper file for openvino."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

from typing import Any, Dict, List

import numpy as np

from otx.api.entities.datasets import DatasetItemEntity
from otx.api.utils.argument_checks import check_input_parameters_type

try:
    from openvino.model_zoo.model_api.adapters import OpenvinoAdapter
    from openvino.model_zoo.model_api.models.model import Model
    from openvino.model_zoo.model_api.models.utils import (
        RESIZE_TYPES,
        Detection,
        InputTransform,
    )
except ImportError as e:
    import warnings

    warnings.warn(f"{e}, ModelAPI was not found.")


@check_input_parameters_type()
def softmax_numpy(x: np.ndarray):
    """Softmax numpy."""
    x = np.exp(x - np.max(x))
    x /= np.sum(x)
    return x


@check_input_parameters_type()
def get_multiclass_predictions(logits: np.ndarray, activate: bool = True):
    """Get multiclass predictions."""
    index = np.argmax(logits)
    if activate:
        logits = softmax_numpy(logits)
    return [(index, logits[index])]


# pylint: disable=too-many-instance-attributes
class OTXOVActionCls(Model):
    """OTX Action Classification model for openvino task."""

    __model__ = "ACTION_CLASSIFICATION"

    def __init__(self, model_adapter: OpenvinoAdapter, configuration=None, preload=False):
        """Image model constructor.

        Calls the `Model` constructor first
        """
        super().__init__(model_adapter, configuration, preload)
        self.image_blob_names, self.image_info_blob_names = self._get_inputs()
        self.image_blob_name = self.image_blob_names[0]
        self.out_layer_name = self._get_outputs()

        _, self.n, self.c, self.t, self.h, self.w = self.inputs[self.image_blob_name].shape
        self.resize = RESIZE_TYPES["standard"]
        self.input_transform = InputTransform(False, None, None)

        # FIXME Below parameters should be changed dynamically from data pipeline config
        self.interval = 4

    def _get_inputs(self):
        image_blob_names, image_info_blob_names = [], []
        for name, metadata in self.inputs.items():
            if len(metadata.shape) == 6:
                image_blob_names.append(name)
            elif len(metadata.shape) == 4:
                image_info_blob_names.append(name)
        return image_blob_names, image_info_blob_names

    def _get_outputs(self):
        layer_name = "logits"
        for name, meta in self.outputs.items():
            if "logits" in meta.names:
                layer_name = name
        return layer_name

    @check_input_parameters_type()
    def preprocess(self, inputs: List[DatasetItemEntity]):
        """Pre-process."""
        meta = {"original_shape": inputs[0].media.numpy.shape}
        frames = []
        for item in inputs:
            frame = item.media.numpy
            frame = self.resize(frame, (self.w, self.h))
            frames.append(frame)
        np_frames = self._reshape(frames)
        dict_inputs = {self.image_blob_name: np_frames}
        meta.update({"resized_shape": np_frames[0].shape})
        return dict_inputs, meta

    @staticmethod
    @check_input_parameters_type()
    def _reshape(inputs: List[np.ndarray]) -> np.ndarray:
        """Reshape(expand, transpose, permute) the input np.ndarray."""
        np_inputs = np.expand_dims(inputs, axis=(0, 1))  # [1, 1, T, H, W, C]
        np_inputs = np_inputs.transpose(0, 1, -1, 2, 3, 4)  # [1, 1, C, T, H, W]
        return np_inputs

    @check_input_parameters_type()
    # pylint: disable=unused-argument
    def postprocess(self, outputs: Dict[str, np.ndarray], meta: Dict[str, Any]):
        """Post-process."""
        logits = outputs[self.out_layer_name].squeeze()
        return get_multiclass_predictions(logits)


class OTXOVActionDet(Model):
    """OTX Action Detection model for openvino task."""

    __model__ = "ACTION_DETECTION"

    def __init__(self, model_adapter: OpenvinoAdapter, configuration=None, preload=False):
        """Image model constructor.

        Calls the `Model` constructor first
        """
        super().__init__(model_adapter, configuration, preload)
        self.image_blob_names = self._get_inputs()
        self.image_blob_name = self.image_blob_names[0]
        self.out_layer_names = self._get_outputs()

        self.n, self.c, self.t, self.h, self.w = self.inputs[self.image_blob_name].shape
        self.resize = RESIZE_TYPES["standard"]
        self.input_transform = InputTransform(False, None, None)

        # FIXME Below parameters should be changed dynamically from data pipeline config
        self.interval = 1
        self.fps = 1

    def _get_inputs(self):
        image_blob_names = []
        for name, metadata in self.inputs.items():
            if len(metadata.shape) == 5:
                image_blob_names.append(name)
        return image_blob_names

    def _get_outputs(self):
        out_names = {}
        for name in self.outputs:
            if "bboxes" in name:
                out_names["bboxes"] = name
            elif "labels" in name:
                out_names["labels"] = name
        return out_names

    @check_input_parameters_type()
    def preprocess(self, inputs: List[DatasetItemEntity]):
        """Pre-process."""
        meta = {"original_shape": inputs[0].media.numpy.shape}
        frames = []
        for item in inputs:
            frame = item.media.numpy
            frame = self.resize(frame, (self.w, self.h))
            frames.append(frame)
        np_frames = self.reshape(frames)
        dict_inputs = {self.image_blob_name: np_frames}
        meta.update({"resized_shape": np_frames.shape})
        return dict_inputs, meta

    @staticmethod
    @check_input_parameters_type()
    def reshape(inputs: List[np.ndarray]) -> np.ndarray:
        """Reshape(expand, transpose, permute) the input np.ndarray."""
        np_inputs = np.expand_dims(inputs, axis=0)  # [1, T, H, W, C]
        np_inputs = np_inputs.transpose(0, -1, 1, 2, 3)  # [1, C, T, H, W]
        return np_inputs

    @check_input_parameters_type()
    def postprocess(self, outputs: Dict[str, np.ndarray], meta: Dict[str, Any]):
        """Post-process."""
        # TODO Support multi label classification
        H, W, _ = meta["original_shape"]
        bboxes = outputs[self.out_layer_names["bboxes"]]
        labels = outputs[self.out_layer_names["labels"]]
        scores = labels[:, 1:].max(axis=1)
        labels = labels[:, 1:].argmax(axis=1)
        results = []
        for bbox, score, label in zip(bboxes, scores, labels):
            bbox *= [W, H, W, H]
            results.append(Detection(*bbox, score, label))
        return results



"""OTX Adapters for action recognition models - mmaction2."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .backbones import register_action_backbones
from .detectors import AVAFastRCNN
from .heads import AVARoIHead

__all__ = ["register_action_backbones", "AVAFastRCNN", "AVARoIHead"]



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmcv.utils import Config


class MockModel(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.conv1 = torch.nn.Conv2d(3, 3, 3)
        self.conv2 = torch.nn.Conv2d(3, 3, 3)
        self.linear = torch.nn.Linear(3, 1)

    def forward(self, img, img_metas=None, **kwargs):
        if isinstance(img, list):
            img = img[0]
        if img.shape[-1] == 3:
            img = img.permute(0, 3, 1, 2)

        x = self.conv1(img)
        x = self.conv2(img)
        x = torch.mean(x, dim=(2, 3))
        x = self.linear(x)
        return x

    def forward_dummy(self, img, img_metas=None, **kwargs):
        if isinstance(img, list):
            img = img[0]
        if img.shape[-1] == 3:
            img = img.permute(0, 3, 1, 2)

        x = self.conv1(img)
        x = self.conv2(img)
        x = torch.mean(x, dim=(2, 3))
        x = self.linear(x)
        return x

    def train_step(self, *args, **kwargs):
        return dict()

    def init_weights(self, *args, **kwargs):
        pass

    def set_step_params(self, *args, **kwargs):
        pass


def create_model(lib="mmcls"):
    if lib == "mmcls":
        from mmcls.models import CLASSIFIERS

        CLASSIFIERS.register_module(MockModel, force=True)
    elif lib == "mmdet":
        from mmdet.models import DETECTORS

        DETECTORS.register_module(MockModel, force=True)
    elif lib == "mmseg":
        from mmseg.models import SEGMENTORS

        SEGMENTORS.register_module(MockModel, force=True)
    else:
        raise ValueError()

    return MockModel()


def create_config():
    config = Config(
        {
            "model": {
                "type": "MockModel",
            },
            "data": {
                "test": {
                    "pipeline": [
                        {"type": "LoadImageFromFile"},
                        {"type": "Normalize", "mean": [0, 0, 0], "std": [1, 1, 1]},
                        {"type": "ImageToTensor", "keys": ["img"]},
                    ]
                }
            },
        }
    )

    return config



"""This module implements the Image entity."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from typing import Optional, Tuple

import cv2
import imagesize
import numpy as np

from otx.api.entities.annotation import Annotation
from otx.api.entities.media import IMedia2DEntity
from otx.api.entities.shapes.rectangle import Rectangle


class Image(IMedia2DEntity):
    """Represents a 2D image.

    The image must be instantiated with either a NumPy array containing the image data
    or a path to an image file.

    Args:
        data (Optional[np.ndarray]): NumPy data.
        file_path (Optional[str]): Path to image file.
    """

    # pylint: disable=too-many-arguments, redefined-builtin
    def __init__(
        self,
        data: Optional[np.ndarray] = None,
        file_path: Optional[str] = None,
    ):
        if (data is None) == (file_path is None):
            raise ValueError("Either path to image file or image data should be provided.")
        self.__data: Optional[np.ndarray] = data
        self.__file_path: Optional[str] = file_path
        self.__height: Optional[int] = None
        self.__width: Optional[int] = None

    def __str__(self):
        """String representation of the image. Returns the image format, name and dimensions."""
        return (
            f"{self.__class__.__name__}"
            f"({self.__file_path if self.__data is None else 'with data'}, "
            f"width={self.width}, height={self.height})"
        )

    def __get_size(self) -> Tuple[int, int]:
        """Returns image size.

        Returns:
            Tuple[int, int]: Image size as a (height, width) tuple.
        """
        if self.__data is not None:
            return self.__data.shape[0], self.__data.shape[1]
        try:
            width, height = imagesize.get(self.__file_path)
            if width <= 0 or height <= 0:
                raise ValueError("Invalide image size")
        except ValueError:
            image = cv2.imread(self.__file_path)
            height, width = image.shape[:2]
        return height, width

    @property
    def numpy(self) -> np.ndarray:
        """Numpy representation of the image.

        For color images the dimensions are (height, width, color) with RGB color channel order.

        Returns:
            np.ndarray: NumPy representation of the image.
        """
        if self.__data is None:
            return cv2.cvtColor(cv2.imread(self.__file_path), cv2.COLOR_BGR2RGB)
        return self.__data

    @numpy.setter
    def numpy(self, value: np.ndarray):
        self.__data = value
        self.__file_path = None
        self.__height, self.__width = self.__get_size()

    def roi_numpy(self, roi: Optional[Annotation] = None) -> np.ndarray:
        """Obtains the numpy representation of the image for a selection region of interest (roi).

        Args:
            roi (Optional[Annotaiton]): The region of interest can be Rectangle in the relative
                coordinate system of the full-annotation.

        Returns:
            np.ndarray: Selected region as numpy array.
        """
        data = self.numpy
        if roi is None:
            return data

        if not isinstance(roi.shape, Rectangle):
            raise ValueError("roi shape is not a Rectangle")

        if data is None:
            raise ValueError("Numpy array is None, and thus cannot be cropped")

        if len(data.shape) < 2:
            raise ValueError("This image is one dimensional, and thus cannot be cropped")

        return roi.shape.crop_numpy_array(data)

    @property
    def height(self) -> int:
        """Returns the height of the image."""
        if self.__height is None:
            self.__height, self.__width = self.__get_size()
        return self.__height

    @property
    def width(self) -> int:
        """Returns the width of the image."""
        if self.__width is None:
            self.__height, self.__width = self.__get_size()
        return self.__width

    @property
    def path(self) -> Optional[str]:
        """Returns the file path of the image."""
        return self.__file_path



# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

try:
    import e2e.fixtures
    from e2e import config  # noqa
    from e2e.conftest_utils import *  # noqa
    from e2e.conftest_utils import pytest_addoption as _e2e_pytest_addoption  # noqa
    from e2e.utils import get_plugins_from_packages

    pytest_plugins = get_plugins_from_packages([e2e])
except ImportError:
    _e2e_pytest_addoption = None
    pass
import pytest

from tests.test_suite.pytest_insertions import (
    get_pytest_plugins_from_otx,
    otx_conftest_insertion,
    otx_pytest_addoption_insertion,
    otx_pytest_generate_tests_insertion,
)
from tests.test_suite.training_tests_common import REALLIFE_USECASE_CONSTANT

pytest_plugins = get_pytest_plugins_from_otx()

otx_conftest_insertion(default_repository_name="otx/training_extensions/external/anomaly")


@pytest.fixture
def otx_test_domain_fx():
    raise NotImplementedError("Please, implement the fixture otx_test_domain_fx in your test file")


@pytest.fixture
def otx_test_scenario_fx(current_test_parameters_fx):
    assert isinstance(current_test_parameters_fx, dict)
    if current_test_parameters_fx.get("usecase") == REALLIFE_USECASE_CONSTANT:
        return "performance"
    else:
        return "integration"


@pytest.fixture(scope="session")
def otx_templates_root_dir_fx():
    import logging
    import os.path as osp

    logger = logging.getLogger(__name__)
    root = osp.dirname(osp.dirname(osp.realpath(__file__)))
    root = osp.realpath(f"{root}/../../otx/algorithms/anomaly/configs")
    logger.debug(f"overloaded otx_templates_root_dir_fx: return {root}")
    return root


@pytest.fixture(scope="session")
def otx_reference_root_dir_fx():
    import logging
    import os.path as osp

    logger = logging.getLogger(__name__)
    root = osp.dirname(osp.realpath(__file__))
    root = f"{root}/reference/"
    logger.debug(f"overloaded otx_reference_root_dir_fx: return {root}")
    return root


# pytest magic
def pytest_generate_tests(metafunc):
    otx_pytest_generate_tests_insertion(metafunc)


def pytest_addoption(parser):
    otx_pytest_addoption_insertion(parser)



"""EfficientNet-B0 for multi-class MMDeploy config."""

_base_ = ["../base/deployments/base_classification_dynamic.py"]

ir_config = dict(
    output_names=["logits"],
)

backend_config = dict(
    model_inputs=[dict(opt_shapes=dict(input=[-1, 3, 224, 224]))],
)



"""Utility functions for attr package.

This module contains utility functions to use with the attr package, concerning for instance parameter validation
or serialization. They are used within the cfg_helper or the configuration elements.
"""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from enum import Enum
from typing import Callable, List, Optional, Type, TypeVar, Union

from attr import Attribute

from otx.api.configuration.elements.configurable_enum import ConfigurableEnum
from otx.api.configuration.elements.parameter_group import ParameterGroup
from otx.api.entities.id import ID

NumericTypeVar = TypeVar("NumericTypeVar", int, float)
SelectableTypeVar = TypeVar("SelectableTypeVar", float, str)
ConfigurableEnumTypeVar = TypeVar("ConfigurableEnumTypeVar", bound=ConfigurableEnum)


def attr_enum_to_str_serializer(
    instance: object,  # pylint: disable=unused-argument
    attribute: Attribute,  # pylint: disable=unused-argument
    value: Union[Enum, str],
) -> str:
    """This function converts Enums to their string representation.

    It is used when converting between yaml and python object representation of the configuration.
    The function signature matches what is expected by the `attr.asdict` value_serializer argument.

    Args:
        instance (object): (Unused) The instance of the class that the attribute is attached to.
        attribute (Attribute): (unused) The attribute that is being serialized.
        value (Union[Enum, str]): The value to serialize.

    Returns:
        str: The string representation of the value.
    """
    if isinstance(value, Enum):
        return str(value)
    return value


def _convert_enum_selectable_value(
    value: Union[str, ConfigurableEnumTypeVar],
    enum_class: Type[ConfigurableEnumTypeVar],
) -> ConfigurableEnumTypeVar:
    """Helper function that converts the input value to an instance of the correct ConfigurableEnum.

    Args:
        value (Union[str, ConfigurableEnumTypeVar]): input value to convert
        enum_class (Type[ConfigurableEnumTypeVar]): Type of the Enum to convert to

    Returns:
        ConfigurableEnumTypeVar: Instance of the correct ConfigurableEnum
    """
    if isinstance(value, str):
        try:
            enum_value = enum_class(value)
        except ValueError as ex:
            raise ValueError(
                f"The value {value} is an invalid option for {enum_class.__name__}. Valid options are:"
                f" {enum_class.get_values()}"
            ) from ex
        return enum_value
    return value


def construct_attr_enum_selectable_converter(
    default_value: ConfigurableEnumTypeVar,
) -> Callable[[Union[str, ConfigurableEnumTypeVar]], ConfigurableEnumTypeVar]:
    """This function converts an input value to the correct instance of the ConfigurableEnum.

    It is used when initializing a selectable parameter.

    Args:
        default_value (ConfigurableEnumTypeVar): Default value for the selectable

    Returns:
        Callable[[Union[str, ConfigurableEnumTypeVar]], ConfigurableEnumTypeVar]:
            Function that converts an input value to the correct instance of the ConfigurableEnum.
    """
    enum_class = type(default_value)

    def attr_convert_enum_selectable_value(value: Union[str, ConfigurableEnumTypeVar]) -> ConfigurableEnumTypeVar:
        """Function that converts an input value to an instance of the appropriate ConfigurableEnum.

        Can be used as a `converter` for attrs.Attributes of type ConfigurableEnum.

        Args:
            value (Union[str, ConfigurableEnumTypeVar]): Value to convert to ConfigurableEnum instance

        Returns:
            ConfigurableEnumTypeVar: Instance of the correct ConfigurableEnum
        """
        return _convert_enum_selectable_value(value, enum_class=enum_class)

    return attr_convert_enum_selectable_value


def construct_attr_enum_selectable_onsetattr(
    default_value: ConfigurableEnumTypeVar,
) -> Callable[[ParameterGroup, Attribute, Union[str, ConfigurableEnumTypeVar]], ConfigurableEnumTypeVar]:
    """This function converts an input value to the correct instance of the ConfigurableEnum.

    It is used when setting a value for a selectable parameter.

    Args:
        default_value (ConfigurableEnumTypeVar): Default value for the enum_selectable.

    Returns:
        Callable[[Union[str, ConfigurableEnumTypeVar]], ConfigurableEnumTypeVar]: Function that converts an input
            value to the correct instance of the ConfigurableEnum.
    """
    enum_class = type(default_value)

    def attr_convert_enum_selectable_value(
        instance: ParameterGroup,  # pylint: disable=unused-argument
        attribute: Attribute,  # pylint: disable=unused-argument
        value: Union[str, ConfigurableEnumTypeVar],
    ) -> ConfigurableEnumTypeVar:
        """Function that converts an input value to an instance of the appropriate ConfigurableEnum.

        Can be used with the `on_setattr` hook of the attrs package.
        """
        return _convert_enum_selectable_value(value, enum_class=enum_class)

    return attr_convert_enum_selectable_value


def construct_attr_value_validator(
    min_value: NumericTypeVar, max_value: NumericTypeVar
) -> Callable[[ParameterGroup, Attribute, NumericTypeVar], None]:
    """Constructs a validator function that is used in the attribute validation of numeric configurable parameters.

    Args:
        min_value (NumericTypeVar): Minimum value for the parameter
        max_value (NumericTypeVar): Maximum value for the parameter

    Returns:
        Callable[[ParameterGroup, Attribute, NumericTypeVar], None]: Function that validates the input value.
    """

    def attr_validate_value(
        instance: ParameterGroup, attribute: Attribute, value: NumericTypeVar
    ):  # pylint: disable=unused-argument
        """This function is used to validate values for numeric ConfigurableParameters."""
        if not min_value <= value <= max_value:
            raise ValueError(f"Invalid value set for {attribute.name}: {value} is out of bounds.")

    return attr_validate_value


def construct_attr_selectable_validator(
    options: List[SelectableTypeVar],
) -> Callable[[ParameterGroup, Attribute, SelectableTypeVar], None]:
    """Constructs a validator function that is used in the attribute validation of selectable configurable parameters.

    Args:
        options (List[SelectableTypeVar]): List of valid options for the parameter.

    Returns:
        Callable[[ParameterGroup, Attribute, SelectableTypeVar], None]: Function that validates the input value.
    """

    def attr_validate_selectable(
        instance: ParameterGroup, attribute: Attribute, value: SelectableTypeVar
    ):  # pylint: disable=unused-argument
        """This function is used to validate values for selectable ConfigurableParameters."""
        if value not in options:
            raise ValueError(
                f"Invalid value set for {attribute.name}: {value} is not a valid option for this " f"parameter."
            )

    return attr_validate_selectable


def convert_string_to_id(id_string: Optional[Union[str, ID]]) -> ID:
    """This function converts an input string representing an ID into an OTX ID object.

    Inputs that are already in the form of an ID are left untouched.

    Args:
        id_string (Optional[Union[str, ID]]): string, ID or None object that should be converted to an ID.

    Returns:
        ID: the input as an instance of ID
    """
    if id_string is None:
        output_id = ID()
    elif isinstance(id_string, str):
        output_id = ID(id_string)
    else:
        output_id = id_string
    return output_id


def attr_strict_int_validator(
    instance: ParameterGroup,  # pylint: disable=unused-argument
    attribute: Attribute,
    value: int,
) -> None:
    """Validates that the value set for an attribute is an integer.

    Args:
        instance (ParameterGroup): (Unused) ParameterGroup to which the attribute belongs
        attribute (Attribute): Attribute for which to validate the value
        value (int): Value to validate

    Raises:
        TypeError: if the value passed to the validator is not an integer
    """
    is_strict_int = isinstance(value, int) and not isinstance(value, bool)
    if not is_strict_int:
        raise TypeError(f"Invalid argument type for {attribute.name}: {value} is not of type 'int'")


def _validate_and_convert_float(value: float) -> Optional[float]:
    """Validate that a value is a float, or a number that can be converted to a float.

    If the value is valid, this method will return the value as float. Otherwise, this
    method returns None

    Args:
        value (float): Value to validate and convert

    Returns:
        Optional[float]: The value as a float, or None if the value is not valid
    """
    valid = True
    if not isinstance(value, (float, int)):
        valid = False
    if isinstance(value, bool):
        valid = False
    if valid:
        return float(value)
    return None


def attr_strict_float_on_setattr(
    instance: ParameterGroup,  # pylint: disable=unused-argument
    attribute: Attribute,
    value: float,
) -> float:
    """Validate that the value set for an attribute is a float, or a number that can be converted to a float.

    Args:
        instance (ParameterGroup): ParameterGroup to which the attribute belongs
        attribute (Attribute): Attribute for which to validate the value
        value (float): Value to validate

    Raises:
        TypeError: if the value passed to the validator is not a float or number

    Returns:
        float: The value as a float
    """
    float_value = _validate_and_convert_float(value)
    if float_value is None:
        raise TypeError(f"Invalid argument type for {attribute.name}: {value} is not of type " f"'float'")
    return float_value


def attr_strict_float_converter(value: float) -> float:
    """Converts a value to float.

    Args:
        value (float): value to convert

    Raises:
        TypeError: if value cannot be converted to float

    Returns:
        Value as float
    """
    float_value = _validate_and_convert_float(value)
    if float_value is None:
        raise TypeError(f"Invalid value passed for parameter. Value {value} of type {type(value)} " f"is not a float.")
    return float_value



"""Tests the methods in the Inference task."""

# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from copy import deepcopy

from otx.algorithms.anomaly.tasks.inference import InferenceTask
from otx.algorithms.anomaly.tasks.train import TrainingTask
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from tests.unit.algorithms.anomaly.helpers.dummy_dataset import get_shapes_dataset
from tests.unit.algorithms.anomaly.helpers.utils import create_task_environment


class TestInferenceTask:
    """Tests the methods in the inference task."""

    def test_inference(self, tmpdir, setup_task_environment):
        """Tests the inference method."""
        root = str(tmpdir.mkdir("anomaly_inference_test"))

        # Get task environment
        setup_task_environment = deepcopy(setup_task_environment)  # since fixture is mutable
        task_environment = setup_task_environment.task_environment
        task_type = setup_task_environment.task_type
        output_model = setup_task_environment.output_model
        dataset = setup_task_environment.dataset

        # 1. Create the training task and get inference results on an untrained model.
        dataset = dataset.get_subset(Subset.VALIDATION)
        train_task = TrainingTask(task_environment, output_path=root)
        dataset = train_task.infer(dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True))
        train_task.save_model(output_model)
        # 2. check if the model is saved correctly
        assert output_model.get_data("weights.pth") is not None  # Should not raise an error

        # 3. Create new task environment and inference task and test inference
        new_dataset: DatasetEntity = get_shapes_dataset(task_type, one_each=True)
        gt_val_dataset = new_dataset.get_subset(Subset.VALIDATION)
        new_task_environment = create_task_environment(gt_val_dataset, task_type)
        # this loads the output model from the previous training task when creating the new InferenceTask
        new_task_environment.model = output_model
        inference_task = InferenceTask(new_task_environment, output_path=root)
        pred_val_dataset = inference_task.infer(
            gt_val_dataset.with_empty_annotations(), InferenceParameters(is_evaluation=True)
        )

        # compare labels with the original validation dataset
        for item1, item2 in zip(dataset, pred_val_dataset):
            assert set([label.name for label in item1.annotation_scene.get_labels()]) == set(
                [label.name for label in item2.annotation_scene.get_labels()]
            )

        # 4. Check whether performance metrics are produced correctly
        # This tests whether the performance metrics are calculated correctly and assigned to the result set
        output_model = ModelEntity(
            gt_val_dataset,
            new_task_environment.get_model_configuration(),
        )
        result_set = ResultSetEntity(
            model=output_model, ground_truth_dataset=gt_val_dataset, prediction_dataset=pred_val_dataset
        )
        inference_task.evaluate(result_set)
        if task_type in (TaskType.ANOMALY_CLASSIFICATION, TaskType.ANOMALY_DETECTION):
            assert result_set.performance.score.name == "f-measure"
        elif task_type == TaskType.ANOMALY_SEGMENTATION:
            assert result_set.performance.score.name == "Dice Average"

        # 5. Check if OpenVINO model can be generated
        inference_task.export(ExportType.OPENVINO, output_model)
        assert output_model.get_data("openvino.bin") is not None  # Should not raise an error



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import torch

from otx.mpa.utils.logger import get_logger

logger = get_logger()


def map_class_names(src_classes, dst_classes):
    """Computes src to dst index mapping

    src2dst[src_idx] = dst_idx
    #  according to class name matching, -1 for non-matched ones
    assert(len(src2dst) == len(src_classes))
    ex)
      src_classes = ['person', 'car', 'tree']
      dst_classes = ['tree', 'person', 'sky', 'ball']
      -> Returns src2dst = [1, -1, 0]
    """
    src2dst = []
    for src_class in src_classes:
        if src_class in dst_classes:
            src2dst.append(dst_classes.index(src_class))
        else:
            src2dst.append(-1)
    return src2dst


def refine_results(results):
    if isinstance(results[0], dict):
        tasks = results[0].keys()
        res_refine = {}
        for task in tasks:
            res_refine[task] = np.concatenate([res[task] for res in results])
    else:
        res_refine = np.vstack(results)
    return res_refine


def extract_anchor_ratio(dataset, num_ratios=5):
    ratio_info = []
    if hasattr(dataset, "dataset"):  # to confirm dataset is wrapped.
        dataset = dataset.dataset
    for ds in dataset:
        ori_shape = ds["img_metas"].data["ori_shape"]
        img_shape = ds["img_metas"].data["img_shape"]
        bboxes = ds["gt_bboxes"].data.numpy()
        for bbox in bboxes:
            w_o = bbox[2] - bbox[0]
            h_o = bbox[3] - bbox[1]
            if w_o > 0.04 * ori_shape[1] and h_o > 0.04 * ori_shape[0]:
                w_i = w_o * img_shape[1] / ori_shape[1]
                h_i = h_o * img_shape[0] / ori_shape[0]
                ratio_info.append(w_i / h_i)
    ratio_info = np.sort(np.array(ratio_info))
    ratio_step = int(len(ratio_info) / num_ratios)
    proposal_ratio = []
    for i in range(num_ratios):
        r = np.mean(ratio_info[i * ratio_step : (i + 1) * ratio_step])
        proposal_ratio.append(r)
    return proposal_ratio


def map_cat_and_cls_as_order(classes, cats):
    cat2label = {}
    cat_ids = []
    for i, cls in enumerate(classes):
        for _, cat in cats.items():
            if cls == cat["name"]:
                cat_id = cat["id"]
                cat_ids.append(cat_id)
                cat2label.update({cat_id: i})
    return cat2label, cat_ids


def unwrap_dataset(dataset):
    times = 1
    target_dataset = dataset
    while hasattr(target_dataset, "dataset"):
        if hasattr(target_dataset, "times"):
            times = target_dataset.times
        target_dataset = target_dataset.dataset
    return target_dataset, times



"""Action Base / Classification / Detection Dataset Adapter."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=invalid-name, too-many-locals, no-member
import os
import os.path as osp
from typing import Any, Dict, List, Optional

from datumaro.components.annotation import AnnotationType
from datumaro.components.annotation import Bbox as DatumaroBbox
from datumaro.components.dataset import Dataset as DatumaroDataset

from otx.api.entities.annotation import Annotation
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import LabelEntity
from otx.api.entities.metadata import MetadataItemEntity, VideoMetadata
from otx.api.entities.subset import Subset
from otx.core.data.adapter.base_dataset_adapter import BaseDatasetAdapter


class ActionBaseDatasetAdapter(BaseDatasetAdapter):
    """BaseDataset Adpater for Action tasks inherited by BaseDatasetAdapter."""

    def _import_dataset(
        self,
        train_data_roots: Optional[str] = None,
        val_data_roots: Optional[str] = None,
        test_data_roots: Optional[str] = None,
        unlabeled_data_roots: Optional[str] = None,
    ) -> Dict[Subset, DatumaroDataset]:
        """Import multiple videos that have CVAT format annotation.

        Args:
            train_data_roots (Optional[str]): Path for training data
            val_data_roots (Optional[str]): Path for validation data
            test_data_roots (Optional[str]): Path for test data
            unlabeled_data_roots (Optional[str]): Path for unlabeled data

        Returns:
            DatumaroDataset: Datumaro Dataset
        """
        dataset = {}
        if train_data_roots is None and test_data_roots is None:
            raise ValueError("At least 1 data_root is needed to train/test.")

        if train_data_roots:
            dataset[Subset.TRAINING] = self._prepare_cvat_pair_data(train_data_roots)
            if val_data_roots:
                dataset[Subset.VALIDATION] = self._prepare_cvat_pair_data(val_data_roots)
        if test_data_roots:
            dataset[Subset.TESTING] = self._prepare_cvat_pair_data(test_data_roots)

        return dataset

    def _prepare_cvat_pair_data(self, path: str) -> List[DatumaroDataset]:
        """Preparing a list of DatumaroDataset."""
        cvat_data_list = []
        for cvat_data in os.listdir(path):
            cvat_data_path = osp.join(path, cvat_data)
            cvat_data_list.append(DatumaroDataset.import_from(cvat_data_path, "cvat"))
        return cvat_data_list

    # pylint: disable=protected-access, too-many-nested-blocks
    def _prepare_label_information(self, datumaro_dataset: dict) -> Dict[str, Any]:
        """Prepare and reorganize the label information for merging multiple video information.

        Description w/ examples:

        [Making overall categories]
        Suppose that video1 has labels=[0, 1, 2] and video2 has labels=[0, 1, 4],
        then the overall label should include all label informations as [0, 1, 2, 4].

        [Reindexing the each label index of multiple video datasets]
        In this case, if the label for 'video1/frame_000.jpg' is 2, then the index of label is set to 2.
        For the case of video2, if the label for 'video2/frame_000.jpg' is 4, then the index of label is set to 2.
        However, Since overall labels are [0, 1, 2, 4], 'video2/frame_000.jpg' should has the label index as 3.

        """
        outputs = {
            "label_entities": [],
        }  # type: dict

        # Making overall categories
        has_EmptyFrame = False
        category_names: List[str] = []  # to check the duplicate case
        subset = list(datumaro_dataset.keys())[0]
        for cvat_data in datumaro_dataset[subset]:
            categories = cvat_data.categories().get(AnnotationType.label, None)
            if categories is not None:
                indices = categories._indices
                for name in indices:
                    if name == "EmptyFrame":
                        has_EmptyFrame = True
                        continue
                    if name not in category_names:
                        category_names.append(name)
        category_names.sort()

        if has_EmptyFrame:
            category_names.append("EmptyFrame")

        # Reindexing the each label index of multiple video datasets
        for subset_data in datumaro_dataset.values():
            for cvat_data in subset_data:
                for cvat_data_item in cvat_data:
                    categories = cvat_data.categories().get(AnnotationType.label, None)
                    if categories is not None:
                        for annotation in cvat_data_item.annotations:
                            ann_name = self.get_ann_name(categories._indices, annotation.label)
                            if ann_name is not None:
                                annotation.label = self.get_ann_label(category_names, ann_name)

        # Generate label_entity list according to overall categories
        outputs["label_entities"] = [
            LabelEntity(name=name, domain=self.domain, is_empty=False, id=ID(index))
            for index, name in enumerate(category_names)
        ]
        return outputs

    @staticmethod
    def get_ann_name(indices: Dict[str, int], label: int) -> Optional[str]:
        """Get action name from index."""
        for _name, _label in indices.items():
            if _label == label:
                return _name
        return None

    @staticmethod
    def get_ann_label(category_names: List[str], ann_name: str) -> Optional[int]:
        """Get action index from name."""
        for idx, name in enumerate(category_names):
            if name == ann_name:
                return idx
        return None

    def get_otx_dataset(self) -> DatasetEntity:
        """Get DatasetEntity.

        Args:
            datumaro_dataset (dict): A Dictionary that includes subset dataset(DatasetEntity)

        Returns:
            DatasetEntity:
        """
        raise NotImplementedError()


class ActionClassificationDatasetAdapter(ActionBaseDatasetAdapter):
    """Action classification adapter inherited by ActionBaseDatasetAdapter and BaseDatasetAdapter."""

    def get_otx_dataset(self) -> DatasetEntity:
        """Convert DatumaroDataset to DatasetEntity for Acion Classification."""
        label_information = self._prepare_label_information(self.dataset)
        self.label_entities = label_information["label_entities"]

        dataset_items: List[DatasetItemEntity] = []
        for subset, subset_data in self.dataset.items():
            for datumaro_items in subset_data:
                for datumaro_item in datumaro_items:
                    image = Image(file_path=datumaro_item.media.path)
                    video_name = datumaro_item.media.path.split("/")[-3]
                    shapes: List[Annotation] = []
                    for annotation in datumaro_item.annotations:
                        if annotation.type == AnnotationType.label:
                            shapes.append(self._get_label_entity(annotation))

                    metadata_item = MetadataItemEntity(
                        data=VideoMetadata(
                            video_id=video_name,
                            frame_idx=int(datumaro_item.media.path.split("/")[-1].split(".")[0].lstrip("0")),
                            is_empty_frame=False,
                        )
                    )

                    dataset_item = DatasetItemEntity(
                        image, self._get_ann_scene_entity(shapes), subset=subset, metadata=[metadata_item]
                    )
                    dataset_items.append(dataset_item)
        return DatasetEntity(items=dataset_items)


class ActionDetectionDatasetAdapter(ActionBaseDatasetAdapter):
    """Action Detection adapter inherited by ActionBaseDatasetAdapter and BaseDatasetAdapter."""

    # pylint: disable=too-many-nested-blocks
    def get_otx_dataset(self) -> DatasetEntity:
        """Convert DatumaroDataset to DatasetEntity for Acion Detection."""
        label_information = self._prepare_label_information(self.dataset)
        self.label_entities = label_information["label_entities"]

        # Detection use index 0 as a background category
        for label_entity in self.label_entities:
            label_entity.id = ID(int(label_entity.id) + 1)

        dataset_items: List[DatasetItemEntity] = []
        for subset, subset_data in self.dataset.items():
            for datumaro_items in subset_data:
                for datumaro_item in datumaro_items:
                    image = Image(file_path=datumaro_item.media.path)
                    video_name = datumaro_item.media.path.split("/")[-3]
                    shapes: List[Annotation] = []
                    is_empty_frame = False
                    for annotation in datumaro_item.annotations:
                        if isinstance(annotation, DatumaroBbox):
                            if self.label_entities[annotation.label].name == "EmptyFrame":
                                is_empty_frame = True
                                shapes.append(self._get_label_entity(annotation))
                            else:
                                shapes.append(self._get_original_bbox_entity(annotation))
                    metadata_item = MetadataItemEntity(
                        data=VideoMetadata(
                            video_id=video_name,
                            frame_idx=int(datumaro_item.media.path.split("/")[-1].split(".")[0].split("_")[-1]),
                            is_empty_frame=is_empty_frame,
                        )
                    )
                    dataset_item = DatasetItemEntity(
                        image, self._get_ann_scene_entity(shapes), subset=subset, metadata=[metadata_item]
                    )
                    dataset_items.append(dataset_item)

        if self.label_entities[-1].name == "EmptyFrame":
            self.label_entities = self.label_entities[:-1]

        return DatasetEntity(items=dataset_items)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.algorithms.detection.tasks import DetectionInferenceTask
from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import ModelConfiguration, ModelEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


class MockDetectionInferenceTask(DetectionInferenceTask):
    def __init__(self):
        pass


class TestInferenceTaskInputParamsValidation:
    @staticmethod
    def model():
        model_configuration = ModelConfiguration(
            configurable_parameters=ConfigurableParameters(header="header", description="description"),
            label_schema=LabelSchemaEntity(),
        )
        return ModelEntity(train_dataset=DatasetEntity(), configuration=model_configuration)

    @e2e_pytest_unit
    def test_otx_detection_inference_task_init_params_validation(self):
        """
        <b>Description:</b>
        Check DetectionInferenceTask object initialization parameters validation

        <b>Input data:</b>
        "task_environment" non-TaskEnvironment object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        DetectionInferenceTask object initialization parameter
        """
        with pytest.raises(ValueError):
            DetectionInferenceTask(task_environment="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_otx_detection_inference_task_infer_params_validation(self):
        """
        <b>Description:</b>
        Check DetectionInferenceTask object "infer" method input parameters validation

        <b>Input data:</b>
        DetectionInferenceTask object. "infer" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "infer" method
        """
        task = MockDetectionInferenceTask()
        correct_values_dict = {
            "dataset": DatasetEntity(),
            "inference_parameters": InferenceParameters(),
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "dataset" parameter
            ("dataset", unexpected_str),
            # Unexpected string is specified as "inference_parameters" parameter
            ("inference_parameters", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=task.infer,
        )

    @e2e_pytest_unit
    def test_otx_detection_inference_task_evaluate_params_validation(self):
        """
        <b>Description:</b>
        Check DetectionInferenceTask object "evaluate" method input parameters validation

        <b>Input data:</b>
        DetectionInferenceTask object. "evaluate" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "evaluate" method
        """
        task = MockDetectionInferenceTask()
        model = self.model()
        result_set = ResultSetEntity(
            model=model,
            ground_truth_dataset=DatasetEntity(),
            prediction_dataset=DatasetEntity(),
        )
        correct_values_dict = {
            "output_result_set": result_set,
            "evaluation_metric": "metric",
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "output_result_set" parameter
            ("output_result_set", unexpected_int),
            # Unexpected integer is specified as "evaluation_metric" parameter
            ("evaluation_metric", unexpected_int),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=task.evaluate,
        )

    @e2e_pytest_unit
    def test_otx_detection_inference_task_export_params_validation(self):
        """
        <b>Description:</b>
        Check DetectionInferenceTask object "export" method input parameters validation

        <b>Input data:</b>
        DetectionInferenceTask object. "export" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "export" method
        """
        task = MockDetectionInferenceTask()
        model = self.model()
        correct_values_dict = {
            "export_type": ExportType.OPENVINO,
            "output_model": model,
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "export_type" parameter
            ("export_type", unexpected_str),
            # Unexpected string is specified as "output_model" parameter
            ("output_model", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=task.export,
        )



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch.nn as nn
from mmcv.runner.checkpoint import _load_checkpoint
from mmdet.models import LOSSES


@LOSSES.register_module()
class L2SPLoss(nn.Module):
    def __init__(self, model, model_ckpt, loss_weight=0.0001):
        """L2-SP regularization loss

        Args:
            model (nn.Module): Input module to regularize
            model_ckpt (str): Starting-point model checkpoint
                Matched params in model would be regularized to be close to starting-point params
            loss_weight (float, optional): Weight of the loss. Defaults to 0.0001
        """
        super().__init__()
        self.model_ckpt = model_ckpt
        self.loss_weight = loss_weight
        print("L2SP loss init!")
        print(f" - starting-point: {model_ckpt}")

        if model_ckpt is None:
            raise ValueError("Model checkpoint path should be provided to enable L2-SP loss!")
        if loss_weight <= 0.0:
            raise ValueError("Loss weight should be a positive value!")

        # Load weights
        src_weights = _load_checkpoint(self.model_ckpt)
        if "state_dict" in src_weights:
            src_weights = src_weights["state_dict"]
        dst_weights = model.named_parameters()

        # Strip 'module.' from weight names if any
        src_weights = {k.replace("module.", ""): v for k, v in src_weights.items()}
        src_weights = {k.replace("model_s.", ""): v for k, v in src_weights.items()}
        # for name in src_weights:
        #    print(name)

        # Match weight name & shape
        self.l2_weights = []
        self.l2sp_weights = []
        for name, p in dst_weights:
            name = name.replace("module.", "")
            if not p.requires_grad:
                print(f"{name}: skip - no grad")
                continue

            if name in src_weights:
                src_w = src_weights[name]
            elif name.replace("backbone.", "") in src_weights:
                # In case of backbone only weights like ImageNet pretrained
                # -> names are stating w/o 'backbone.'
                src_w = src_weights[name.replace("backbone.", "")]
            else:
                self.l2_weights.append(p)
                print(f"{name}: l2 - new param")
                continue

            if p.shape != src_w.shape:
                # Same name but w/ different shape
                self.l2_weights.append(p)
                print(f"{name}: l2 - diff shape ({p.shape} vs {src_w.shape}")
                continue

            src_w.requires_grad = False
            self.l2sp_weights.append((p, src_w))
            print(f"{name}: l2sp")

    def forward(self, **kwargs):
        """Forward function.

        Returns:
            torch.Tensor: The calculated loss
        """

        # loss = torch.tensor(0.0, requires_grad=True)
        loss = 0.0
        # L2 loss
        for w in self.l2_weights:
            loss += (w**2).sum()
        # L2-SP loss
        for w, w0 in self.l2sp_weights:
            loss += ((w - w0.to(w)) ** 2).sum()

        return self.loss_weight * loss



_base_ = ["./data.py", "./pipelines/coco_inst_seg_pipeline.py"]

__dataset_type = "CocoDataset"
__data_root = "data/coco/"

__train_pipeline = {{_base_.train_pipeline}}
__test_pipeline = {{_base_.test_pipeline}}

__samples_per_gpu = 4

data = dict(
    samples_per_gpu=__samples_per_gpu,
    workers_per_gpu=2,
    train=dict(
        type=__dataset_type,
        ann_file=__data_root + "annotations/instances_train2017.json",
        img_prefix=__data_root + "train2017/",
        pipeline=__train_pipeline,
    ),
    val=dict(
        type=__dataset_type,
        ann_file=__data_root + "annotations/instances_val2017.json",
        img_prefix=__data_root + "val2017/",
        test_mode=True,
        pipeline=__test_pipeline,
    ),
    test=dict(
        type=__dataset_type,
        ann_file=__data_root + "annotations/instances_val2017.json",
        img_prefix=__data_root + "val2017/",
        test_mode=True,
        pipeline=__test_pipeline,
    ),
)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from functools import partial

import numpy as np
import torch
from mmcls.core import EvalHook
from mmcls.datasets import build_dataloader
from mmcls.datasets.pipelines import Compose
from mmcv.utils import Config, ConfigDict, get_logger
from torch.optim import SGD

import otx.algorithms.common.adapters.mmcv.nncf.patches  # noqa: F401
from otx.algorithms.common.adapters.mmcv.nncf.hooks import CompressionHook
from otx.algorithms.common.adapters.mmcv.nncf.runners import AccuracyAwareRunner
from otx.algorithms.common.adapters.mmcv.nncf.utils import (
    get_fake_input,
    model_eval,
    wrap_nncf_model,
)
from otx.algorithms.common.adapters.mmcv.utils import build_data_parallel
from otx.algorithms.common.adapters.nncf.patches import nncf_trace_context


def create_model(lib="mmcls"):
    class MockModel(torch.nn.Module):
        def __init__(self, *args, **kwargs):
            super().__init__()
            self.conv1 = torch.nn.Conv2d(3, 3, 3)
            self.conv2 = torch.nn.Conv2d(3, 3, 3)
            self.linear = torch.nn.Linear(3, 1)

        def forward(self, img, img_metas=None, **kwargs):
            if isinstance(img, list):
                img = img[0]
            if img.shape[-1] == 3:
                img = img.permute(0, 3, 1, 2)

            x = self.conv1(img)
            x = self.conv2(img)
            x = torch.mean(x, dim=(2, 3))
            x = self.linear(x)
            return x

        def forward_dummy(self, img, img_metas=None, **kwargs):
            if isinstance(img, list):
                img = img[0]
            if img.shape[-1] == 3:
                img = img.permute(0, 3, 1, 2)

            x = self.conv1(img)
            x = self.conv2(img)
            x = torch.mean(x, dim=(2, 3))
            x = self.linear(x)
            return x

        def train_step(self, *args, **kwargs):
            return dict()

        def init_weights(self, *args, **kwargs):
            pass

        def set_step_params(self, *args, **kwargs):
            pass

    MockModel.nncf_trace_context = nncf_trace_context

    if lib == "mmcls":
        from mmcls.models import CLASSIFIERS

        CLASSIFIERS.register_module(MockModel, force=True)
    elif lib == "mmdet":
        from mmdet.models import DETECTORS

        DETECTORS.register_module(MockModel, force=True)
    elif lib == "mmseg":
        from mmseg.models import SEGMENTORS

        SEGMENTORS.register_module(MockModel, force=True)
    else:
        raise ValueError()

    return MockModel()


def create_dataset(pipeline=None, lib="mmcls"):
    class MockDataset(torch.utils.data.Dataset):
        def __init__(self, pipeline, *args, **kwargs):
            super().__init__()
            self.dataset = np.zeros((10, 128, 128, 3), dtype=np.float32)

            if isinstance(pipeline, list):
                if lib == "mmcls":
                    from mmcls.datasets.pipelines import Compose
                elif lib == "mmdet":
                    from mmdet.datasets.pipelines import Compose
                elif lib == "mmseg":
                    from mmseg.datasets.pipelines import Compose
                else:
                    raise ValueError()
                pipeline = Compose(pipeline)
            self.pipeline = pipeline

        def __getitem__(self, idx):
            if self.pipeline:
                return self.pipeline(
                    dict(
                        img=self.dataset[idx],
                        filename=None,
                        ori_filename=None,
                        img_fields=["img"],
                        img_shape=self.dataset[idx].shape,
                        ori_shape=self.dataset[idx].shape,
                    )
                )
            return self.dataset[idx]

        def __len__(self):
            return len(self.dataset)

        def evaluate(self, model_output, **kwargs):
            #  if isinstance(model_output, list):
            #      model_output = torch.cat(model_output)
            #  return {"accuracy": model_output.mean()}
            return {"accuracy": torch.tensor(0.9)}

    if lib == "mmcls":
        from mmcls.datasets import DATASETS
    elif lib == "mmdet":
        from mmdet.datasets import DATASETS
    elif lib == "mmseg":
        from mmseg.datasets import DATASETS
    else:
        raise ValueError()

    DATASETS.register_module(MockDataset, force=True)
    return MockDataset(pipeline)


def create_dataloader(config=None):
    if config is None:
        config = create_config()
    pipeline = Compose(config.data.val.pipeline)
    mock_dataset = create_dataset(pipeline)
    dataloader = build_dataloader(mock_dataset, samples_per_gpu=2, workers_per_gpu=1)
    return dataloader


def create_config(lib="mmcls"):
    config = Config(
        {
            "gpu_ids": [0],
            "evaluation": {"metric": "accuracy"},
            "data": {
                "samples_per_gpu": 2,
                "workers_per_gpu": 0,
                "persistent_workers": False,
                "shuffle": False,
            },
            "model": {
                "type": "MockModel",
            },
            "nncf_config": {
                "input_info": {"sample_size": (1, 3, 128, 128)},
                "target_metric_name": "accuracy",
                "compression": [
                    {
                        "algorithm": "quantization",
                        "preset": "mixed",
                        "initializer": {
                            "range": {"num_init_samples": 10},
                            "batchnorm_adaptation": {"num_bn_adaptation_samples": 10},
                        },
                    }
                ],
                "accuracy_aware_training": {"params": {"maximal_total_epochs": 5, "mode": "early_exit"}},
            },
            "runner": {
                "type": "AccuracyAwareRunner",
                "nncf_config": {
                    "input_info": {"sample_size": (1, 3, 128, 128)},
                    "target_metric_name": "accuracy",
                    "compression": [
                        {
                            "algorithm": "quantization",
                            "preset": "mixed",
                            "initializer": {
                                "range": {"num_init_samples": 10},
                                "batchnorm_adaptation": {"num_bn_adaptation_samples": 10},
                            },
                        }
                    ],
                    "accuracy_aware_training": {"params": {"maximal_total_epochs": 5, "mode": "early_exit"}},
                },
            },
        }
    )

    for subset in ("train", "test", "val"):
        if lib == "mmcls":
            config.data[subset] = ConfigDict(
                {
                    "type": "MockDataset",
                    "pipeline": [
                        {"type": "Resize", "size": (50, 50)},
                        {"type": "Normalize", "mean": [0, 0, 0], "std": [1, 1, 1]},
                        {"type": "ImageToTensor", "keys": ["img"]},
                        {
                            "type": "Collect",
                            "keys": ["img"],
                            "meta_keys": [
                                "filename",
                                "ori_filename",
                                "ori_shape",
                                "img_shape",
                                "img_norm_cfg",
                            ],
                        },
                    ],
                }
            )
        else:
            config.data[subset] = ConfigDict(
                {
                    "type": "MockDataset",
                    "pipeline": [
                        {"type": "Resize", "img_scale": (50, 50)},
                        {"type": "Normalize", "mean": [0, 0, 0], "std": [1, 1, 1]},
                        {"type": "ImageToTensor", "keys": ["img"]},
                        {
                            "type": "Collect",
                            "keys": ["img"],
                            "meta_keys": [
                                "filename",
                                "ori_filename",
                                "ori_shape",
                                "img_shape",
                                "img_norm_cfg",
                            ],
                        },
                    ],
                }
            )
    return config


def create_eval_fn():
    def evaluate_fn(model, loader, *args, **kwargs):
        out = []
        for data in loader:
            out.append(torch.sigmoid(model(**data)))
        return torch.cat(out)

    return evaluate_fn


def create_nncf_model():
    mock_model = create_model()
    mock_config = create_config()
    mock_eval_fn = create_eval_fn()
    dataloader = create_dataloader()

    mock_config = create_config()
    pipeline = Compose(mock_config.data.val.pipeline)
    get_fake_input_fn = partial(get_fake_input, pipeline)

    model_eval_fn = partial(
        model_eval,
        config=mock_config,
        val_dataloader=dataloader,
        evaluate_fn=mock_eval_fn,
        distributed=False,
    )

    ctrl, model = wrap_nncf_model(
        mock_config,
        mock_model,
        model_eval_fn=model_eval_fn,
        get_fake_input_fn=get_fake_input_fn,
        dataloader_for_init=dataloader,
        is_accuracy_aware=True,
    )
    return ctrl, model


def create_nncf_runner(work_dir):
    mock_config = create_config()
    dataloader = create_dataloader()
    ctrl, model = create_nncf_model()

    runner = AccuracyAwareRunner(
        build_data_parallel(model, mock_config),
        logger=get_logger("mmcv"),
        work_dir=work_dir,
        nncf_config=mock_config["nncf_config"],
        optimizer=SGD(model.parameters(), lr=0.01),
    )
    runner.register_hook(
        EvalHook(
            dataloader,
            save_best="accuracy",
            file_client_args={"backend": "disk"},
            priority="LOW",
        )
    )
    runner.register_hook(CompressionHook(ctrl))
    return runner
